from __future__ import annotations

from pydantic import BaseModel, Field, field_validator, ConfigDict


class ExtraParamsSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    energy_kev: float | None = Field(default=None, description="Energy in keV")
    x_sample_pixel_size_m: float | None = Field(
        default=None, description="X sample pixel size in meters"
    )
    y_sample_pixel_size_m: float | None = Field(
        default=None, description="Y sample pixel size in meters"
    )
    x_detector_pixel_size_m: float | None = Field(
        default=None, description="X detector pixel size in meters"
    )
    y_detector_pixel_size_m: float | None = Field(
        default=None, description="Y detector pixel size in meters"
    )
    detector_sample_distance_m: float | None = Field(
        default=None, description="Detector sample distance in meters"
    )
    source_sample_distance_m: float | None = Field(
        default=None, description="Source sample distance in meters"
    )

    @field_validator(
        "energy_kev",
        "x_sample_pixel_size_m",
        "y_sample_pixel_size_m",
        "x_detector_pixel_size_m",
        "y_detector_pixel_size_m",
        "detector_sample_distance_m",
        "source_sample_distance_m",
        mode="plain",
    )
    @classmethod
    def cast_extra_params_to_scalar_value_or_None(cls, value: str) -> float | None:
        if value in ("", None, "None"):
            return None
        else:
            return float(value)



# coding: utf-8

import os

from nxtomomill.app.h52nx import main
from nxtomomill.tests.utils.bliss import MockBlissAcquisition


def test_h52nx_application(tmp_path):
    """test nxtomomill h52nx input_file output_file --single-file"""
    folder = tmp_path / "test_h52nx_application"
    folder.mkdir()
    nx_file_path = os.path.join(folder, "acquisition.nx")

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=5,
        n_flats=5,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="pcolinux",
    )

    h5_file_path = bliss_mock.samples[0].sample_file
    assert not os.path.exists(nx_file_path), "outputfile exists already"
    main(["h52nx", h5_file_path, nx_file_path, "--single-file"])
    assert os.path.exists(nx_file_path), "outputfile doesn't exists"



# coding: utf-8

from __future__ import annotations


from nxtomomill.utils.io import deprecated, deprecated_warning
from nxtomomill.models.fluo2nx import Fluo2nxModel as _TomoFluoConfig
from nxtomomill.models.fluo2nx import (
    generate_default_fluo_config as _generate_default_fluo_config,
)

__all__ = ["TomoFluoConfig", "generate_default_fluo_config"]


class TomoFluoConfig(_TomoFluoConfig):
    def __init__(self, *args, **kwargs):
        deprecated_warning(
            type_="class",
            name="TomoFluoConfig",
            replacement="nxtomomill.models.fluo2nx.Fluo2nxModel",
            reason="moved to models module",
            since_version="2.0",
        )
        super().__init__(*args, **kwargs)


@deprecated(
    reason="moved to from nxtomomill.models.fluo2nx module", since_version="2.0"
)
def generate_default_fluo_config(*args, **kwargs):
    return _generate_default_fluo_config(*args, **kwargs)



from __future__ import annotations

from .multitomo import MultiTomoAcquisition


class BackAndForthAcquisition(MultiTomoAcquisition):
    """

        Back and forth acquisition is very similar to multi-tomo.

        In multi-tomo, the acquisition of projections begins when the rotation speed has reached an 'optimal' value.
    value.

        In back-and-forth acquisition, projections will be acquired in one rotation direction, then in the opposite direction, and so on.

        From the nxtomomill perspective, the constraints and processing remain the same.
        Nevertheless for clarity (and possible future evolution) a dedicated class has been created.
    """



# coding: utf-8

"""
Application to convert a tomo dataset written in edf into and hdf5/nexus file.

.. program-output:: nxtomomill edf2nx --help

For a complete tutorial you can have a look at :ref:`edf2nxtutorial`
"""

import argparse
import logging

from nxtomomill import converter
from nxtomomill.models.utils import convert_str_to_tuple
from nxtomomill.io.config.edfconfig import TomoEDFConfig
from nxtomomill.settings import Tomo
from tqdm import tqdm

logging.basicConfig(level=logging.INFO)

fabio_logger = logging.getLogger("fabio.edfimage")
fabio_logger.setLevel(logging.ERROR)


def main(argv):
    """ """
    parser = argparse.ArgumentParser(
        description="convert data acquired as "
        "edf to hdf5 - nexus "
        "compliant file format."
    )
    parser.add_argument("scan_path", help="folder containing the edf files", nargs="?")
    parser.add_argument("output_file", help="foutput .h5 file", nargs="?")
    parser.add_argument(
        "--dataset-basename",
        "--file-prefix",
        default=None,
        help="file prefix to be used to deduce projections",
    )
    parser.add_argument(
        "--info-file",
        default=None,
        help=".info file containing acquisition information (ScanRange, Energy, TOMO_N...)",
    )
    parser.add_argument(
        "--config",
        "--configuration-file",
        "--configuration",
        default=None,
        help="file containing the full configuration to convert from SPEC-EDF to bliss to nexus. "
        "Default configuration file can be created from `nxtomomill edf-config` command",
    )
    parser.add_argument(
        "--delete-edf",
        "--delete-edf-source",
        "--delete-edf-source-files",
        default=False,
        action="store_true",
        help="if you are duplicating data (default behavior) you can ask to delete used edf files to free space",
    )
    parser.add_argument(
        "--output-checks",
        default=tuple(),
        help="Define list of check to be run once the conversion is finished (raise an error if check fails). This is done before removing edf if asked. So if check fails source edf files won't be removed",
    )
    parser.add_argument(
        "--use-existing-angles",
        "--no-angle-recalculcation",
        default=None,
        action="store_true",
        help=f"By default the angle will be recomputed to have equally spaced angles from the min and the max angles. If this option is set then angles defined in edf headers ({','.join(Tomo.EDF.ROT_ANGLE)}) will be used.",
    )
    options = parser.parse_args(argv[1:])
    config = TomoEDFConfig()
    if options.config is not None:
        config = config.from_cfg_file(options.config)

    check_input = {
        "dataset basename": (options.dataset_basename, config.dataset_basename),
        "scan path": (options.scan_path, config.input_folder),
        "output file": (options.output_file, config.output_file),
        "info file": (options.info_file, config.dataset_info_file),
    }
    for input_name, (opt_value, config_value) in check_input.items():
        if (
            opt_value is not None
            and config_value is not None
            and opt_value != config_value
        ):
            raise ValueError(
                f"two values for {input_name} are provided from arguments and configuration file ({opt_value, config_value})"
            )

    if options.dataset_basename is not None:
        config.dataset_basename = options.dataset_basename
    if options.info_file is not None:
        config.dataset_info_file = options.info_file
    if options.scan_path is not None:
        config.input_folder = options.scan_path
    if options.output_file is not None:
        config.output_file = options.output_file
    config.delete_edf_source_files = options.delete_edf
    if options.use_existing_angles is not None:
        config.force_angle_calculation = not options.use_existing_angles
    config.output_checks = convert_str_to_tuple(options.output_checks)

    assert isinstance(
        config.output_checks, tuple
    ), f"config.output_checks is expected to be a tuple. Gets {type(config.output_checks)}"
    converter.from_edf_to_nx(
        configuration=config,
        progress=tqdm(desc="edf2nx", bar_format="{l_bar}{bar}{postfix}"),
    )



"""module with various utils like modifying image_key values, checking if and HDF5 entry looks like an NXtomo entry..."""

from .frameappender import FrameAppender  # noqa F401
from .utils import FileExtension  # noqa F401
from .utils import add_dark_flat_nx_file  # noqa F401
from .utils import change_image_key_control  # noqa F401
from .utils import embed_url  # noqa F401
from .utils import get_file_name  # noqa F401
from .utils import get_tuple_of_keys_from_cmd  # noqa F401
from .utils import is_nx_tomo_entry  # noqa F401

# expose ImageKey from nxtomo
from nxtomo.nxobject.nxdetector import ImageKey  # noqa F401



import datetime

import pytest
from tomoscan.io import HDF5File
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan

from nxtomomill.app.zstages2nxs import _convert_bliss2nx
from nxtomomill.tests.utils.bliss import MockBlissAcquisition as _MockBlissAcquisition
from nxtomomill.utils.flat_reducer import flat_reducer


class MockBlissAcquisition(_MockBlissAcquisition):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for sample in self.samples:
            # append machine current to the scan
            with HDF5File(sample.sample_file, mode="a") as h5f:
                for i_bliss_scan in range(
                    1, 1 + self._n_scan_per_sequence + self._n_darks + self._n_flats
                ):
                    node_name = f"{i_bliss_scan}.1"
                    h5f[f"{node_name}/instrument/machine/current"] = [
                        12.3,
                    ]
                    h5f[f"{node_name}/instrument/machine/current"].attrs["units"] = "mA"
                    if f"{node_name}/start_time" in h5f:
                        del h5f[f"{node_name}/start_time"]
                    h5f[f"{node_name}/start_time"] = str(datetime.datetime.now())
                    if f"{node_name}/end_time" in h5f:
                        del h5f[f"{node_name}/end_time"]
                    h5f[f"{node_name}/end_time"] = str(
                        datetime.datetime.now() + datetime.timedelta(minutes=10)
                    )


@pytest.mark.parametrize("stage_scan_have_flats", (True, False))
@pytest.mark.parametrize("save_intermediated", (True, False))
@pytest.mark.parametrize("reuse_intermediated", (True, False))
def test_flat_reducer(
    tmp_path, stage_scan_have_flats, save_intermediated, reuse_intermediated
):
    """
    test execution of `flat_reducer` function.
    This function is going through a set of bliss acquisition to create corresponding NXtomo and add them dark and flats obtained from
    'reference' acquisition (converting projections to reduced flats)
    """
    raw_data_dir = tmp_path / "raw"
    n_stage = 5
    # create bliss scan to be converted to NXtomo
    acqu = MockBlissAcquisition(
        n_sample=n_stage,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=2,
        n_flats=1 if stage_scan_have_flats else 0,
        output_dir=raw_data_dir,
        file_name_z_fill=4,
    )
    stages_bliss_file = [sample.sample_file for sample in acqu.samples]

    # scan for flat at start
    ref_start = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=2,
        n_flats=0,
        output_dir=raw_data_dir,
        file_name_prefix="REF_B_0000",
    )
    bliss_ref_start = ref_start.samples[0].sample_file
    nx_tomo_ref_start = bliss_ref_start.replace(".h5", ".nx")
    _convert_bliss2nx(
        bliss_ref_start,
        nx_tomo_ref_start,
    )
    # scan for flat at end
    ref_end = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=2,
        n_flats=0,
        output_dir=raw_data_dir,
        file_name_prefix="REF_E_0000",
    )
    bliss_ref_end = ref_end.samples[0].sample_file
    nx_tomo_ref_end = bliss_ref_end.replace(".h5", ".nx")
    _convert_bliss2nx(
        bliss_ref_end,
        nx_tomo_ref_end,
    )

    for i_stage, bliss_file_name in enumerate(stages_bliss_file):
        mixing_factor = (i_stage + 1) / (n_stage)

        # convert from .h5 to .nx
        output_nx_file = bliss_file_name.replace(".h5", ".nx")
        _convert_bliss2nx(
            bliss_file_name,
            output_nx_file,
        )

        # apply fly reducer
        flat_reducer(
            scan_filename=output_nx_file,
            ref_start_filename=nx_tomo_ref_start,
            ref_end_filename=nx_tomo_ref_end,
            mixing_factor=mixing_factor,
            save_intermediated=save_intermediated,
            reuse_intermediated=reuse_intermediated,
        )
        # make sure there is some reduced dark and flat for the nx

        i_stage_scan = NXtomoScan(output_nx_file, "entry0000")
        i_stage_reduced_flats = i_stage_scan.load_reduced_flats()
        i_stage_reduced_darks = i_stage_scan.load_reduced_darks()
        assert len(i_stage_reduced_flats) == 1
        assert len(i_stage_reduced_darks) == 1



# coding: utf-8
"""
module to convert from (bliss) .h5 to (nexus tomo compliant) .nx
"""

from __future__ import annotations

import logging
import os
import sys

import h5py

from tqdm import tqdm

from silx.io.url import DataUrl
from silx.io.utils import open as open_hdf5
from tomoscan.io import HDF5File
from tomoscan.utils.io import filter_esrf_mounting_points

from nxtomomill.converter.hdf5.acquisition.utils import group_series
from nxtomomill.converter.baseconverter import BaseConverter
from nxtomomill.converter.hdf5.acquisition.baseacquisition import _ask_for_file_removal
from nxtomomill.converter.hdf5.acquisition.multitomo import MultiTomoAcquisition
from nxtomomill.models.h52nx._acquisitionstep import AcquisitionStep
from nxtomomill.models.h52nx import H52nxModel

from .acquisition.baseacquisition import BaseAcquisition
from .acquisition.standardacquisition import StandardAcquisition
from .acquisition.utils import get_bliss_scan_type
from .acquisition.zseriesacquisition import (
    ZSeriesBaseAcquisition,
)
from .post_processing.dark_flat_copy import ZSeriesDarkFlatCopy
from .acquisitionConstructor import (
    _AcquisitionConstructorFromTitles,
    _AcquisitionConstructorFromUrls,
)

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass
# import that should be removed when h5_to_nx will be removed
from nxtomomill.converter.hdf5.utils import H5FileKeys, H5ScanTitles
from nxtomomill.settings import Tomo
from nxtomomill.utils._os import abspath

H5_ROT_ANGLE_KEYS = Tomo.H5.ROT_ANGLE_KEYS
H5_VALID_CAMERA_NAMES = Tomo.H5.VALID_CAMERA_NAMES
H5_SAMPLE_X_KEYS = Tomo.H5.SAMPLE_X_KEYS
H5_SAMPLE_Y_KEYS = Tomo.H5.SAMPLE_Y_KEYS
H5_TRANSLATION_Z_KEYS = Tomo.H5.TRANSLATION_Z_KEYS
H5_ALIGNMENT_TITLES = Tomo.H5.ALIGNMENT_TITLES
H5_ACQ_EXPO_TIME_KEYS = Tomo.H5.ACQ_EXPO_TIME_KEYS
H5_SAMPLE_X_PIXEL_SIZE = Tomo.H5.SAMPLE_X_PIXEL_SIZE_KEYS
H5_SAMPLE_Y_PIXEL_SIZE = Tomo.H5.SAMPLE_Y_PIXEL_SIZE_KEYS
H5_DETECTOR_X_PIXEL_SIZE = Tomo.H5.DETECTOR_X_PIXEL_SIZE_KEYS
H5_DETECTOR_Y_PIXEL_SIZE = Tomo.H5.DETECTOR_Y_PIXEL_SIZE_KEYS
H5_DARK_TITLES = Tomo.H5.DARK_TITLES
H5_INIT_TITLES = Tomo.H5.INIT_TITLES
H5_MULTITOMO_INIT_TITLES = Tomo.H5.MULTITOMO_INIT_TITLES
H5_BACK_AND_FORTH_INIT_TITLES = Tomo.H5.BACK_AND_FORTH_INIT_TITLES
H5_ZSERIE_INIT_TITLES = Tomo.H5.ZSERIE_INIT_TITLES
H5_PROJ_TITLES = Tomo.H5.PROJ_TITLES
H5_FLAT_TITLES = Tomo.H5.FLAT_TITLES
H5_REF_TITLES = H5_FLAT_TITLES
H5_TRANSLATION_Y_KEYS = Tomo.H5.TRANSLATION_Y_KEYS
H5_DIODE_KEYS = Tomo.H5.DIODE_KEYS

# deprecated variables
H5_PCOTOMO_INIT_TITLES = H5_MULTITOMO_INIT_TITLES


DEFAULT_SCAN_TITLES = H5ScanTitles(
    H5_INIT_TITLES,
    H5_ZSERIE_INIT_TITLES,
    H5_MULTITOMO_INIT_TITLES,
    H5_BACK_AND_FORTH_INIT_TITLES,
    H5_DARK_TITLES,
    H5_FLAT_TITLES,
    H5_PROJ_TITLES,
    H5_ALIGNMENT_TITLES,
)


DEFAULT_H5_KEYS = H5FileKeys(
    H5_ACQ_EXPO_TIME_KEYS,
    H5_ROT_ANGLE_KEYS,
    H5_VALID_CAMERA_NAMES,
    H5_SAMPLE_X_KEYS,
    H5_SAMPLE_Y_KEYS,
    H5_TRANSLATION_Z_KEYS,
    H5_TRANSLATION_Y_KEYS,
    H5_SAMPLE_X_PIXEL_SIZE,
    H5_SAMPLE_Y_PIXEL_SIZE,
    H5_DETECTOR_X_PIXEL_SIZE,
    H5_DETECTOR_Y_PIXEL_SIZE,
    H5_DIODE_KEYS,
)


_logger = logging.getLogger(__name__)


class _H5ToNxConverter(BaseConverter):
    """
    Class used to convert a HDF5Config to one or several NXTomoEntry.

    :param configuration: configuration for the translation. such as the
                          input and output file, keys...
    :param input_callback: possible callback in case of missing information
    :param progress: progress bar to be updated if provided
    :param detector_sel_callback: callback for the detector selection if any

    Conversion is a two step process:

    step 1: preprocessing
    * insure configuration is valid and that we don't have "unsafe" or
      "opposite" request / rules
    * normalize input URL (complete data_file if not provided)
    * copy some frame group if requested
    * create instances of BaseAcquisition classes that will be used to write
      NXTomo entries
    * handle z series specific case

    step 2: write NXTomo entries to the output file
    """

    def __init__(
        self,
        configuration: H52nxModel,
        input_callback=None,
        progress: tqdm | None = None,
        detector_sel_callback=None,
    ):
        if not isinstance(configuration, H52nxModel):
            raise TypeError(
                f"configuration should be an instance of HDFConfig not {type(configuration)}"
            )

        self._configuration = configuration
        self._progress = progress
        self._input_callback = input_callback
        self._detector_sel_callback = detector_sel_callback
        self._acquisitions = []
        self._entries_created = []
        self._z_series_v2_v3: list[list[ZSeriesBaseAcquisition]] = []
        # bliss z-series for version 2 and 3. Can be used for post-processing
        self.preprocess()

    @property
    def configuration(self):
        return self._configuration

    @property
    def progress(self):
        return self._progress

    @property
    def input_callback(self):
        return self._input_callback

    @property
    def detector_sel_callback(self):
        return self._detector_sel_callback

    @property
    def entries_created(self) -> tuple:
        """tuple of entries created. Each element is provided as
        (output_file, entry)"""
        return tuple(self._entries_created)

    @property
    def acquisitions(self):
        return self._acquisitions

    def preprocess(self):
        # clean path
        if self._configuration.input_file is not None:
            self._configuration.input_file = filter_esrf_mounting_points(
                self._configuration.input_file
            )
        if self._configuration.output_file is not None:
            self._configuration.output_file = filter_esrf_mounting_points(
                self._configuration.output_file
            )
        self._preprocess_urls()
        self._check_conversion_is_possible()

        if self.configuration.is_using_titles:
            self._convert_entries_and_sub_entries_to_urls()
            acquisition_builder = _AcquisitionConstructorFromTitles(
                configuration=self.configuration,
                progress=self.progress,
                detector_sel_callback=self.detector_sel_callback,
            )
            self._acquisitions = acquisition_builder.build_sequence()
        else:
            self.configuration.clear_entries_and_subentries()
            acquisition_builder = _AcquisitionConstructorFromUrls(
                configuration=self.configuration,
                progress=self.progress,
                detector_sel_callback=self.detector_sel_callback,
            )
            self._acquisitions = acquisition_builder.build_sequence()
        self._z_series_v2_v3 = self._handle_zseries()

    def _handle_zseries(self):
        # for z series we have a "master" acquisition of type
        # ZSeriesBaseAcquisition. But this is used only to build
        # the acquisition sequence. To write we use the z series
        # "sub_acquisitions" which are instances of "StandardAcquisition"
        acquisitions = []
        z_series_v2_to_v3 = []

        for acquisition in self.acquisitions:
            if isinstance(acquisition, StandardAcquisition):
                acquisitions.append(acquisition)
            elif isinstance(acquisition, ZSeriesBaseAcquisition):
                sub_acquisitions = acquisition.get_standard_sub_acquisitions()
                acquisitions.extend(sub_acquisitions)
                for sub_acquisition in sub_acquisitions:
                    z_series_v2_to_v3 = group_series(
                        acquisition=sub_acquisition, list_of_series=z_series_v2_to_v3
                    )
            else:
                raise TypeError(f"Acquisition type {type(acquisition)} not handled")
        self._acquisitions = acquisitions
        return z_series_v2_to_v3

    def convert(self):
        mess_conversion = f"start conversion from {self.configuration.input_file} to {self.configuration.output_file}"
        if self.progress is not None:
            # in the case we want to print progress
            sys.stdout.write(mess_conversion)
            sys.stdout.flush()
        else:
            _logger.info(mess_conversion)

        self._entries_created = self.write()
        return self._entries_created

    def _ignore_sub_entry(self, sub_entry_url: DataUrl | None):
        """
        :return: True if the provided sub_entry should be ignored
        """
        if sub_entry_url is None:
            return False
        if not isinstance(sub_entry_url, DataUrl):
            raise TypeError(
                f"sub_entry_url is expected to be a DataUrl not {type(sub_entry_url)}"
            )
        if self.configuration.sub_entries_to_ignore is None:
            return False

        sub_entry_fp = sub_entry_url.file_path()
        sub_entry_dp = sub_entry_url.data_path()
        for entry in self.configuration.sub_entries_to_ignore:
            assert isinstance(entry, DataUrl)
            if entry.file_path() == sub_entry_fp and entry.data_path() == sub_entry_dp:
                return True
        return False

    def write(self):
        res = []

        acq_str = [str(acq) for acq in self.acquisitions]
        acq_str.insert(
            0, f"parsing finished. {len(self.acquisitions)} acquisitions found"
        )
        _logger.debug("\n   - ".join(acq_str))

        if len(self.acquisitions) == 0:
            _logger.warning(
                "No valid acquisitions have been found. Most likely "
                "no init titles have been found. You can provide more valid entries from CLI or configuration file."
            )

        if self.progress is not None:
            progress_write = tqdm(desc="write NXtomos")
            progress_write.total = len(self.acquisitions)
        else:
            progress_write = None

        # write nx_tomo per acquisition
        has_single_acquisition_in_file = len(self.acquisitions) == 1 and isinstance(
            self.acquisitions, MultiTomoAcquisition
        )
        divide_into_sub_files = not (
            self.configuration.single_file is False and has_single_acquisition_in_file
        )

        acquisition_to_nxtomo: dict[ZSeriesBaseAcquisition, tuple[str] | None] = {}
        for acquisition in self.acquisitions:
            if self._ignore_sub_entry(acquisition.root_url):
                acquisition_to_nxtomo[acquisition] = None
                continue

            try:
                new_entries = acquisition.write_as_nxtomo(
                    shift_entry=acquisition.start_index,
                    input_file_path=self.configuration.input_file,
                    request_input=self.configuration.request_input,
                    input_callback=self.input_callback,
                    divide_into_sub_files=divide_into_sub_files,
                )
            except Exception as e:
                if self.configuration.raises_error:
                    raise e
                else:
                    root_location = (
                        acquisition.root_url.path()
                        if acquisition.root_url is not None
                        else ""
                    )
                    _logger.error(
                        f"Fail to convert '{root_location}' sequence. Error is {str(e)}",
                        exc_info=e,
                    )
                    acquisition_to_nxtomo[acquisition] = None
            else:
                res.extend(new_entries)
                acquisition_to_nxtomo[acquisition] = new_entries
            if progress_write is not None:
                progress_write.update()

        # post processing on nxtomos
        for series in self._z_series_v2_v3:
            self._post_process_series(series, acquisition_to_nxtomo)

        # if we created one file per entry then create a master file with link to those entries
        if (
            self.configuration.single_file is False and divide_into_sub_files
        ) and not self.configuration.no_master_file:
            _logger.info(f"create link in {self.configuration.output_file}")
            for en_output_file, entry in res:
                with HDF5File(self.configuration.output_file, "a") as master_file:
                    link_file = os.path.relpath(
                        filter_esrf_mounting_points(en_output_file),
                        filter_esrf_mounting_points(
                            os.path.dirname(self.configuration.output_file)
                        ),
                    )
                    master_file[entry] = h5py.ExternalLink(link_file, entry)

        return tuple(res)

    def _check_conversion_is_possible(self):
        """Insure minimalistic information are provided"""
        if self.configuration.is_using_titles:
            if self.configuration.input_file is None:
                raise ValueError("input file should be provided")
            if not os.path.isfile(self.configuration.input_file):
                raise ValueError(
                    f"Given input file does not exists: {self.configuration.input_file}"
                )
            if not h5py.is_hdf5(self.configuration.input_file):
                raise ValueError("Given input file is not an hdf5 file")

        if self.configuration.input_file == self.configuration.output_file:
            raise ValueError("input and output file are the same")

        output_file = self.configuration.output_file
        dir_name = os.path.dirname(abspath(output_file))
        if not os.path.exists(dir_name):
            os.makedirs(os.path.dirname(abspath(output_file)))
        elif os.path.exists(output_file):
            if self.configuration.overwrite is True:
                _logger.warning(f"{output_file} will be removed")
                _logger.info(f"remove {output_file}")
                os.remove(output_file)
            elif not _ask_for_file_removal(output_file):
                raise OSError(f"unable to overwrite {output_file}, exit")
            else:
                os.remove(output_file)
        if not os.access(dir_name, os.W_OK):
            raise OSError(f"You don't have rights to write on {dir_name}")

    def _convert_entries_and_sub_entries_to_urls(self):
        if len(self.configuration.entries) > 0:
            urls = self.configuration.entries
            entries = self._upgrade_urls(
                urls=urls, input_file=self.configuration.input_file
            )
            self.configuration.entries = entries
        if self.configuration.sub_entries_to_ignore is not None:
            urls = self.configuration.sub_entries_to_ignore
            entries = self._upgrade_urls(
                urls=urls, input_file=self.configuration.input_file
            )
            self.configuration.sub_entries_to_ignore = entries

    def _preprocess_urls(self):
        """
        Update darks, flats, projections and alignments urls if
        no file path is provided
        """
        self.configuration.data_scans = self._upgrade_frame_grp_urls(
            frame_grps=self.configuration.data_scans,
            input_file=self.configuration.input_file,
            default_copy=self.configuration.default_data_copy,
        )

    def _post_process_series(
        self,
        series: list[BaseAcquisition],
        acquisition_to_nxtomo: dict[BaseAcquisition, tuple | None],
    ):
        dark_flat_copy = ZSeriesDarkFlatCopy(
            series=series, acquisition_to_nxtomo=acquisition_to_nxtomo
        )
        dark_flat_copy.run()

    @staticmethod
    def _upgarde_url(url: DataUrl, input_file: str) -> DataUrl:
        if url is not None and url.file_path() in (None, ""):
            if input_file in (None, str):
                raise ValueError(
                    f"file_path for url {url.path()} is not provided and no input_file provided either."
                )
            else:
                return DataUrl(
                    file_path=input_file,
                    scheme="silx",
                    data_slice=url.data_slice(),
                    data_path=url.data_path(),
                )
        else:
            return url

    @staticmethod
    def _upgrade_frame_grp_urls(
        frame_grps: tuple, input_file: str | None, default_copy: bool
    ) -> tuple:
        """
        Upgrade all Frame Group DataUrl which did not contain a file_path to
         reference the input_file
        """
        if input_file is not None and not h5py.is_hdf5(input_file):
            raise ValueError(f"{input_file} is not a h5py file")
        for frame_grp in frame_grps:
            frame_grp.url = _H5ToNxConverter._upgarde_url(frame_grp.url, input_file)
            if frame_grp.copy_data is None:
                frame_grp.copy_data = default_copy
        return frame_grps

    @staticmethod
    def _upgrade_urls(urls: tuple, input_file: str | None) -> tuple:
        """
        Upgrade all DataUrl which did not contain a file_path to reference
        the input_file
        """
        if input_file is not None and not h5py.is_hdf5(input_file):
            raise ValueError(f"{input_file} is not a h5py file")
        return tuple([_H5ToNxConverter._upgarde_url(url, input_file) for url in urls])


def from_h5_to_nx(
    configuration: H52nxModel,
    input_callback=None,
    progress: tqdm | None = None,
    detector_sel_callback=None,
):
    """
    convert a bliss file to a set of NXtomo

    :param configuration: configuration for the translation. such as the
                          input and output file, keys...
    :param input_callback: possible callback in case of missing information
    :param progress: progress bar to be updated if provided
    :param detector_sel_callback: callback for the detector selection if any
    :return: tuple of created NXtomo as (output_file, data_path)
    """
    converter = _H5ToNxConverter(
        configuration=configuration,
        input_callback=input_callback,
        progress=progress,
        detector_sel_callback=detector_sel_callback,
    )
    return converter.convert()


def get_bliss_tomo_entries(input_file_path: str, configuration: H52nxModel):
    """.
    Return the set of entries at root that match bliss entries.
    Used by tomwer for example.

    :param input_file_path: path of the file to browse
    :param configuration: configuration of the conversion. This way user can define title to be used or frame groups

    Warning: entries can be external links (in the case of the file being a proposal file)
    """
    if not isinstance(configuration, H52nxModel):
        raise TypeError("configuration is expected to be a HDF5Config")

    with open_hdf5(input_file_path) as h5d:
        acquisitions = []

        for group_name in h5d.keys():
            _logger.debug(f"parse {group_name}")
            entry = h5d[group_name]
            # improve handling of External (this is the case of proposal files)
            if isinstance(h5d.get(group_name, getlink=True), h5py.ExternalLink):
                external_link = h5d.get(group_name, getlink=True)
                file_path = external_link.filename
                data_path = external_link.path
            else:
                file_path = input_file_path
                data_path = entry.name
                if not data_path.startswith("/"):
                    data_path = "/" + data_path
            url = DataUrl(file_path=file_path, data_path=data_path)
            if configuration.is_using_titles:
                # if use title take the ones corresponding to init
                entry_type = get_bliss_scan_type(url=url, configuration=configuration)
                if entry_type is AcquisitionStep.INITIALIZATION:
                    acquisitions.append(group_name)
            else:
                # check if the entry fit one of the data_scans
                # with an init status
                possible_url_file_path = [
                    abspath(url.file_path()),
                    url.file_path(),
                ]
                if configuration.output_file not in ("", None):
                    possible_url_file_path.append(
                        os.path.relpath(
                            url.file_path(), os.path.dirname(configuration.output_file)
                        )
                    )
                for frame_grp in configuration.data_scans:
                    if frame_grp.frame_type is AcquisitionStep.INITIALIZATION:
                        if (
                            frame_grp.url.file_path() in possible_url_file_path
                            and frame_grp.data_path() == url.data_path()
                        ):
                            acquisitions.append(entry.name)

        return acquisitions



from __future__ import annotations

from pydantic import BaseModel, Field, ConfigDict, field_validator
from nxtomomill.models.utils import filter_str_def, convert_str_to_bool


class SampleSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    sample_name: str | None = Field(default=None, description="Name of the sample")
    force_angle_calculation: bool = Field(
        default=True,
        description="Determines the method for obtaining rotation angles."
        "Options:"
        "* True: Compute from scan range: Uses `numpy.linspace` to generate rotation angles based on the provided scan range."
        "* False: Load from `.edf` header: Attempts to read the rotation angles directly from the `.edf` file header.",
    )
    angle_calculation_endpoint: bool = Field(
        default=False,
        description="Specifies the endpoint behavior for `numpy.linspace` when calculating rotation angles."
        "",
    )
    angle_calculation_rev_neg_scan_range: bool = Field(
        default=True,
        description="Inverts the rotation angle values when the `ScanRange` is negative.",
    )

    @field_validator(
        "sample_name",
        mode="plain",
    )
    @classmethod
    def cast_to_sample_name(cls, value: str | None) -> str | None:
        return filter_str_def(value)

    @field_validator(
        "angle_calculation_endpoint",
        "force_angle_calculation",
        "angle_calculation_rev_neg_scan_range",
        mode="plain",
    )
    @classmethod
    def cast_to_bool(cls, value: bool | str) -> bool:
        return convert_str_to_bool(value)



"""
module to convert from `dxchange <https://dxchange.readthedocs.io/en/latest/>`_ (.dx) to  `NXtomo <https://manual.nexusformat.org/classes/applications/NXtomo.html>`_ (.nx)
"""

from .dxfileconverter import from_dx_to_nx  # noqa F401



# coding: utf-8

import os
import pytest
import logging

from configparser import ConfigParser
import importlib.resources as importlib_resources

from nxtomo.nxobject.nxsource import ProbeType, SourceType

from nxtomomill.models.fluo2nx import Fluo2nxModel, generate_default_fluo_config
from nxtomomill.tests.resources import fluo2nx_config_files as _fluo2nx_config_files
from nxtomomill.utils import FileExtension


def test_moving_to_pydantic_default_config_file():
    """
    test that the default nxtomomill config generated before moving to pydantic is properly interpreted.
    This ignores comments.
    """
    resources = importlib_resources.files(_fluo2nx_config_files.__name__)
    ref_file = os.path.join(resources, "default_file_before_moving_to_pydantic.cfg")
    assert os.path.exists(ref_file)
    txt_parser = ConfigParser(allow_no_value=True)
    txt_parser.read(ref_file)

    old_config = Fluo2nxModel(
        general_section=txt_parser["GENERAL_SECTION"],
        source_section=txt_parser["SOURCE_SECTION"],
    ).to_nested_model()
    old_dict = {key.upper(): value for key, value in old_config.model_dump().items()}

    new_dict = generate_default_fluo_config()
    assert new_dict == old_dict


def test_TomoFluoConfig_default_config(tmp_path):
    """insure default configuration generation works"""
    file_path = os.path.join(tmp_path, "config.cfg")
    assert not os.path.exists(file_path)
    Fluo2nxModel().to_cfg_file(file_path=file_path)
    assert os.path.exists(file_path)


def test_moving_to_pydantic_modified_config_file():
    """
    test that an old nxtomomill fluo-config file generated before moving to pydantic is properly interpreted.
    This ignores any comments.
    """
    resources = importlib_resources.files(_fluo2nx_config_files.__name__)
    ref_file = os.path.join(resources, "fully_modified_file_before_moving.cfg")
    assert os.path.exists(ref_file)

    config = Fluo2nxModel.from_cfg_file(file_path=ref_file)
    assert config.input_folder == "my_folder"
    assert config.output_file == "output_file.nx"
    assert config.dimension == 2
    assert config.detector_names == ("my_detector",)
    assert config.overwrite is True
    assert config.file_extension == FileExtension.H5
    assert config.dataset_basename == "dataset_basename"
    assert config.dataset_info_file == "dataset_info_file.info"
    assert config.log_level == logging.ERROR
    # note: 'title' is part of the configuration file but was never used in the processing
    assert config.patterns_to_ignores == ("_ignore_",)
    assert config.duplicate_data is False
    # note: 'external_link_path is part of the configuration file but was never used in the processing
    assert config.instrument_name == "my_instrument"
    assert config.source_name == "my_source"
    assert config.source_type == SourceType.PULSED_REACTOR_NEUTRON_SOURCE
    assert config.source_probe == ProbeType.POSITRON


def test_TomoFluoConfig_setters():
    """test the different setters and getter of the Fluo2nxModel"""
    config = Fluo2nxModel()

    # try to test a new attribut (insure class is frozeen)
    with pytest.raises(ValueError):
        config.new_attrs = "toto"

    # test general section setters
    with pytest.raises(ValueError):
        config.input_folder = 12.0
    config.input_folder = "my_folder"
    with pytest.raises(ValueError):
        config.output_file = 12.0
    config.output_file = "my_nx.nx"

    config.file_extension = ".nx"

    with pytest.raises(ValueError):
        config.detectors = 12.0
    config.detector_names = ()
    config.detector_names = ("test_",)

    with pytest.raises(ValueError):
        config.dataset_basename = 12.0
    config.dataset_basename = None
    config.dataset_basename = "test_"

    config.overwrite = True

    # test source setters
    config.instrument_name = None
    config.instrument_name = "BMXX"
    with pytest.raises(ValueError):
        config.instrument_name = 12.3

    config.source_name = None
    config.source_name = "ESRF"
    with pytest.raises(ValueError):
        config.source_name = 12.2
    config.source_name = "XFEL"

    config.source_type = None
    config.source_type = SourceType.FIXED_TUBE_X_RAY.value
    config.source_type = SourceType.FIXED_TUBE_X_RAY
    with pytest.raises(ValueError):
        config.source_type = "trsts"
    with pytest.raises(ValueError):
        config.source_type = 123

    config.source_probe = None
    config.source_probe = ProbeType.ELECTRON
    config.source_probe = "positron"
    with pytest.raises(ValueError):
        config.source_probe = 123

    config_dict = config.to_dict()
    general_section_dict = config_dict["GENERAL_SECTION"]
    assert general_section_dict["input_folder"] == "my_folder"
    assert general_section_dict["output_file"] == "my_nx.nx"
    assert general_section_dict["file_extension"] == ".nx"
    assert general_section_dict["overwrite"] is True
    assert general_section_dict["dataset_basename"] == "test_"

    source_section_dict = config_dict["SOURCE_SECTION"]
    assert source_section_dict["instrument_name"] == "BMXX"
    assert source_section_dict["source_name"] == "XFEL"
    assert source_section_dict["source_type"] == SourceType.FIXED_TUBE_X_RAY.value



"""a set of applications and tools around the `NXtomo <https://manual.nexusformat.org/classes/applications/NXtomo.html>`_ format defined by the `NeXus community <https://manual.nexusformat.org/index.html>`_"""

from .version import version as __version

__version__ = __version



# coding: utf-8

"""
Utils related to bliss-HDF5
"""

from collections import namedtuple
import os
from pathlib import Path
from nxtomomill.utils._os import abspath

H5FileKeys = namedtuple(
    "H5FileKeys",
    [
        "acq_expo_time_keys",
        "rot_angle_keys",
        "valid_camera_names",
        "sample_x_keys",
        "sample_y_keys",
        "translation_z_keys",
        "translation_y_keys",
        "x_sample_pixel_size",
        "y_sample_pixel_size",
        "x_detector_pixel_size",
        "y_detector_pixel_size",
        "diode_keys",
    ],
)


H5ScanTitles = namedtuple(
    "H5ScanTitles",
    [
        "init_titles",
        "zseries_init_titles",
        "multitomo_init_titles",
        "back_and_forth_init_titles",
        "dark_titles",
        "flat_titles",
        "projection_titles",
        "align_titles",
    ],
)

PROCESSED_DATA_DIR_NAME = "PROCESSED_DATA"
RAW_DATA_DIR_NAME = "RAW_DATA"


def get_default_output_file(input_file: str, output_file_extension: str = ".nx") -> str:
    """
    Policy: look for any 'RAW_DATA' in file directory. If find any (before any 'PROCESSED_DATA' directory) replace it "RAW_DATA".
    Then replace input_file by the expected file_extension and make sure the output file is different than the input file. Else append _nxtomo to it.

    :param input_file: file to be converted from bliss to NXtomo
    :param output_file_extension:
    :return: default output file according to policy
    """
    if isinstance(input_file, Path):
        input_file = str(input_file)
    if not isinstance(input_file, str):
        raise TypeError(
            f"input_file is expected to be an instance of str. {type(input_file)} provided"
        )
    if not isinstance(output_file_extension, str):
        raise TypeError("output_file_extension is expected to be a str")
    if not output_file_extension.startswith("."):
        output_file_extension = "." + output_file_extension

    input_file = abspath(input_file)
    input_file_no_ext, _ = os.path.splitext(input_file)

    def from_raw_data_path_to_process_data_path(file_path: str):
        split_path = file_path.split(os.sep)
        # reverse it to find the lower level value of '_RAW_DATA_DIR_NAME' if by any 'chance' has several in the path
        # in this case this is most likely what we want
        split_path = split_path[::-1]
        # check if already contain in a "PROCESSED_DATA" directory
        try:
            index_processed_data = split_path.index(PROCESSED_DATA_DIR_NAME)
        except ValueError:
            index_processed_data = None

        try:
            index_raw_data = split_path.index(RAW_DATA_DIR_NAME)
        except ValueError:
            # if the value is not in the list
            pass
        else:
            if index_processed_data is None or index_raw_data < index_processed_data:
                # make sure we are not already in a 'PROCESSED_DATA' directory. Not sure it will never happen but safer
                split_path[index_raw_data] = PROCESSED_DATA_DIR_NAME

        # reorder path to original
        split_path = split_path[::-1]
        return os.sep.join(split_path)

    output_path = from_raw_data_path_to_process_data_path(input_file_no_ext)
    output_file = output_path + output_file_extension
    if output_file == input_file:
        # to be safer if the default output file is the same as the input file (if the input file has a .nx extension and not in any 'RAw_DATA' directory)
        return output_path + "_nxtomo" + output_file_extension
    else:
        return output_file



# coding: utf-8

"""
module to define some converter utils function
"""

import h5py
import h5py._hl.selections as selection
from silx.io.url import DataUrl
from silx.io.utils import open as open_hdf5

__all__ = ["from_data_url_to_virtual_source", "from_virtual_source_to_data_url"]


def from_data_url_to_virtual_source(url: DataUrl) -> tuple:
    """
    :param url: url to be converted to a virtual source. It must target a 2D detector
    :return: (h5py.VirtualSource, tuple(shape of the virtual source), numpy.drype: type of the dataset associated with the virtual source)
    """
    if not isinstance(url, DataUrl):
        raise TypeError(
            f"url is expected to be an instance of DataUrl and not {type(url)}"
        )

    with open_hdf5(url.file_path()) as o_h5s:
        original_data_shape = o_h5s[url.data_path()].shape
        data_type = o_h5s[url.data_path()].dtype
        if len(original_data_shape) == 2:
            original_data_shape = (
                1,
                original_data_shape[0],
                original_data_shape[1],
            )

        vs_shape = original_data_shape
        if url.data_slice() is not None:
            vs_shape = (
                url.data_slice().stop - url.data_slice().start,
                original_data_shape[-2],
                original_data_shape[-1],
            )

    vs = h5py.VirtualSource(
        url.file_path(), url.data_path(), shape=vs_shape, dtype=data_type
    )

    if url.data_slice() is not None:
        vs.sel = selection.select(original_data_shape, url.data_slice())
    return vs, vs_shape, data_type


def from_virtual_source_to_data_url(vs: h5py.VirtualSource) -> DataUrl:
    if not isinstance(vs, h5py.VirtualSource):
        raise TypeError(
            f"vs is expected to be an instance of h5py.VirtualSorce and not {type(vs)}"
        )
    url = DataUrl(file_path=vs.path, data_path=vs.name, scheme="silx")
    return url



# coding: utf-8

"""
Application to convert a fluo-tomo dataset, after PyMCA (https://www.silx.org/doc/PyMca/dev/index.html) fit, into an hdf5/nexus file.

.. program-output:: nxtomomill fluo2nx --help

"""

import argparse
import logging
import os

from nxtomomill import utils
from nxtomomill import converter
from nxtomomill.io.config.fluoconfig import TomoFluoConfig

from tqdm import tqdm

logging.basicConfig(level=logging.INFO)


def main(argv):
    """ """
    parser = argparse.ArgumentParser(
        description="Converts fluo-tomo data (after PyMca fit) "
        "to hdf5 - nexus compliant file format."
    )
    parser.add_argument(
        "scan_path",
        help="Path to the folder containing the raw data folder and the fluofit/ subfolder.",
        nargs=1,
    )
    parser.add_argument(
        "output_file",
        help="File produced by the converter. '.nx' extension recommended.",
        nargs=1,
    )
    parser.add_argument(
        "dataset_basename",
        help="In 2D, the exact full name of the folder. In 3D, the folder name prefix (the program will search for folders named <prefix>_projection_XXX where XXX is a nmber.)",
        nargs=1,
    )
    parser.add_argument(
        "--detectors",
        nargs="*",
        default=list(),
        help="Define a list of (real or virtual) detector names used for the exp (space separated values - no comma). E.g. 'falcon xmap'. If not specified, all detectors are processed.",
    )
    parser.add_argument(
        "--dimension",
        help="2 for 2D XRFCT, 3 for 3D XRFCT. Default is 3.",
        default=3,
    )
    parser.add_argument(
        "--info-file",
        default=None,
        help=".info file containing acquisition information (ScanRange, Energy, TOMO_N...)",
    )
    parser.add_argument(
        "--config",
        "--configuration-file",
        "--configuration",
        default=None,
        help="file containing the full configuration to convert from (PyMCA-computed) fluo projections to nexus. "
        "Default configuration file can be created from `nxtomomill fluo-config` command",
    )

    parser.add_argument(
        "--mode",
        default="newfile",
        help="What to do if file exists:"
        "- 'newfile': program will fail if file already exists."
        "- 'overwrite': program will overwrite existing file or existing entries in file.",
    )

    options = parser.parse_args(argv[1:])
    config = TomoFluoConfig()
    if options.config is not None:
        config = config.from_cfg_file(options.config)

    check_input = {
        "dataset basename": (options.dataset_basename, config.dataset_basename),
        "scan path": (options.scan_path, config.input_folder),
        "output file": (options.output_file, config.output_file),
        "info file": (options.info_file, config.dataset_info_file),
    }
    for input_name, (opt_value, config_value) in check_input.items():
        if (
            opt_value is not None
            and config_value is not None
            and opt_value != config_value
        ):
            raise ValueError(
                f"two values for {input_name} are provided from arguments and configuration file ({opt_value, config_value})"
            )

    if options.dataset_basename is not None:
        config.dataset_basename = options.dataset_basename[0]
    if options.info_file is not None:
        config.dataset_info_file = options.info_file
    if options.scan_path is not None:
        config.input_folder = options.scan_path[0]
    if options.output_file is not None:
        config.output_file = options.output_file[0]
    if options.detectors is not None:
        config.detector_names = options.detectors
    config.dimension = options.dimension

    fileout_h5 = utils.get_file_name(
        file_name=config.output_file,
        extension=config.file_extension,
        check=True,
    )
    if os.path.exists(fileout_h5):
        if options.mode == "newfile":
            raise FileExistsError(
                f"The file {fileout_h5} already exists. Please remove it before running the program."
            )
        elif options.mode == "overwrite":
            config.overwrite = True
        else:
            raise RuntimeError(
                f"The value entered for --mode is not valid. Should be either 'newfile' or 'overwrite'. {options.mode} was given."
            )

    converter.from_fluo_to_nx(
        configuration=config,
        progress=tqdm("fluo2nx"),
    )



"""**deprecated**, use `nxtomo <https://tomotools.gitlab-pages.esrf.fr/nxtomo/>`_ module instead"""

from .nxdetector import NXdetector  # noqa F401
from .nxobject import NXobject  # noqa F401
from .nxsample import NXsample  # noqa F401
from .nxsource import NXsource  # noqa F401
from .nxtomo import NXtomo  # noqa F401
from .utils import concatenate  # noqa F401



# coding: utf-8

import logging

import pytest

from nxtomomill.converter.fluo.fluoscan import FluoTomoScan3D
from nxtomomill.tests.datasets import GitlabDataset

logging.disable(logging.INFO)


@pytest.fixture(scope="class")
def fluodata3D(request):
    cls = request.cls
    cls.scan_dir = GitlabDataset.get_dataset("fluo_datasets")
    cls.dataset_basename = "CP1_XRD_insitu_top_ft_100nm"
    cls.scan = FluoTomoScan3D(
        scan=cls.scan_dir,
        dataset_basename=cls.dataset_basename,
        detectors=(),
    )


@pytest.mark.usefixtures("fluodata3D")
class TestFluo3D:
    def test_all_detectors(self):
        assert (
            len(self.scan.el_lines) == 14  # pylint: disable=E1101
        ), f"Number of elements found should be 14 and is {len(self.scan.el_lines)}."  # pylint: disable=E1101
        assert set(self.scan.detectors) == set(  # pylint: disable=E1101
            ["falcon", "weighted", "xmap"]
        ), f"There should be 3 'detectors' (xmap, falcon and weighted), {len(self.detectors)} were found."  # pylint: disable=E1101

    def test_one_detector(self):
        scan = FluoTomoScan3D(
            scan=self.scan_dir,  # pylint: disable=E1101
            dataset_basename=self.dataset_basename,  # pylint: disable=E1101
            detectors=("xmap",),
        )
        assert (
            len(scan.el_lines) == 14
        ), f"Number of elements found should be 14 and is {len(scan.el_lines)}."
        assert (
            len(scan.detectors) == 1
        ), f"There should be 1 detector (xmap), {len(scan.detectors)} were found."

        # One ghost detector (no corresponding files)
        # test general section setters
        with pytest.raises(ValueError):
            scan = FluoTomoScan3D(
                scan=self.scan_dir,  # pylint: disable=E1101
                dataset_basename=self.dataset_basename,  # pylint: disable=E1101
                detectors=("toto",),
            )

    def test_load_data(self):
        data = self.scan.load_data("xmap", "Ti", 0)  # pylint: disable=E1101
        assert data.shape == (2, 51, 280)

    def test_load_energy_and_pixel_size(self):
        assert self.scan.pixel_size == 1.3e-6  # pylint: disable=E1101
        assert self.scan.energy == 17.1  # pylint: disable=E1101



# coding: utf-8

"""
contains the FrameGroup
"""

from enum import Enum as _Enum
from nxtomo.nxobject.nxdetector import ImageKey

__all__ = [
    "AcquisitionStep",
]


class AcquisitionStep(_Enum):
    # Warning: order of acquisition step should be same as H5ScanTitles
    INITIALIZATION = "initialization"
    DARK = "darks"
    FLAT = "flats"
    PROJECTION = "projections"
    ALIGNMENT = "alignment projections"

    @classmethod
    def from_value(cls, value):
        if isinstance(value, str):
            value = value.lower()
            if value in ("init", "initialization"):
                value = AcquisitionStep.INITIALIZATION
            elif value in ("dark", "darks"):
                value = AcquisitionStep.DARK
            elif value in ("reference", "flat", "flats", "ref", "refs", "references"):
                value = AcquisitionStep.FLAT
            elif value in ("proj", "projection", "projs", "projections"):
                value = AcquisitionStep.PROJECTION
            elif value in (
                "alignment",
                "alignments",
                "alignment projection",
                "alignment projections",
            ):
                value = AcquisitionStep.ALIGNMENT

        return AcquisitionStep(value)

    def to_image_key(self):
        if self is AcquisitionStep.PROJECTION:
            return ImageKey.PROJECTION
        elif self is AcquisitionStep.ALIGNMENT:
            return ImageKey.PROJECTION
        elif self is AcquisitionStep.DARK:
            return ImageKey.DARK_FIELD
        elif self is AcquisitionStep.FLAT:
            return ImageKey.FLAT_FIELD
        else:
            raise ValueError(f"The step {self.value} does not fit any AcquisitionStep")

    def to_image_key_control(self):
        if self is AcquisitionStep.PROJECTION:
            return ImageKey.PROJECTION
        elif self is AcquisitionStep.ALIGNMENT:
            return ImageKey.ALIGNMENT
        elif self is AcquisitionStep.DARK:
            return ImageKey.DARK_FIELD
        elif self is AcquisitionStep.FLAT:
            return ImageKey.FLAT_FIELD
        else:
            raise ValueError(f"The step {self.value} does not fit any AcquisitionStep")



from nxtomomill.utils.io import deprecated_warning
from nxtomo.nxobject.nxinstrument import *  # noqa F401

deprecated_warning(
    type_="Module",
    name="nxtomomill.nexus.nxinstrument",
    reason="dedicated project created",
    replacement="nxtomo.nxobject.nxinstrument",
    since_version=1.0,
)



from __future__ import annotations

import re
import logging

from pydantic import BaseModel, Field, field_validator, ConfigDict, field_serializer
from nxtomomill.models.h52nx.FrameGroup import FrameGroup
from nxtomomill.models.utils import remove_parenthesis_or_brackets
from nxtomomill.utils.io import deprecated

_logger = logging.getLogger(__name__)


class FrameTypeSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    data_scans: tuple[FrameGroup, ...] = Field(
        default=tuple(),
        description="List of scans to be converted."
        "Frame type should be provided for each scan. Expected format is:"
        " * `frame_type` (mandatory): values can be `projection`, `flat`, `dark`, `alignment` or `init`."
        " * `entry` (mandatory): DataUrl with path to the scan to integrate. If the scan is contained in the input_file then you can only provide path/name of the scan."
        " * `copy` (optional): you can provide a different behavior for the this scan (should we duplicate data or not)",
    )
    default_data_copy: bool = Field(
        default=False,
        description="Duplicate data inside the input file or create a relative link",
    )

    @property
    @deprecated(reason="renamed", since_version="2.0", replacement="data_scans")
    def data_frame_grps(self):
        return self.data_scans

    @data_frame_grps.setter
    @deprecated(reason="renamed", since_version="2.0", replacement="data_scans")
    def data_frame_grps(self, values) -> None:
        self.data_scans = values

    @field_validator(
        "data_scans",
        mode="plain",
    )
    @classmethod
    def cast_to_data_scan(
        cls, value: str | tuple[FrameGroup, ...]
    ) -> tuple[FrameGroup, ...]:
        if isinstance(value, tuple):
            return value
        return convert_str_to_frame_grp(value)

    @field_serializer(
        "data_scans",
        when_used="always",
    )
    @classmethod
    def serialize_data_scans(cls, urls: tuple[FrameGroup, ...]) -> str:
        if len(urls) == 0:
            return ""
        else:
            urls_str = ",\n".join([f"{frame_group}" for frame_group in urls])
            return f"""(
{urls_str}
)
            """


def convert_str_to_frame_grp(input_str: str) -> tuple[FrameGroup, ...]:
    """
    Convert a list such as:

    .. code-block:: text

        urls = (
            (frame_type=dark, entry="silx:///file.h5?data_path=/dark", copy=True),
            (frame_type=flat, entry="silx:///file.h5?data_path=/flat"),
            (frame_type=projection, entry="silx:///file.h5?data_path=/flat"),
            (frame_type=projection, entry="silx:///file.h5?data_path=/flat"),
        )

    to a tuple of FrameGroup
    """
    result = []
    if not isinstance(input_str, str):
        raise TypeError(
            f"input_str should be an instance of str. Got {type(input_str)}"
        )
    # remove spaces at the beginning and at the end
    input_str = input_str.replace("\n", "")
    input_str = input_str.lstrip(" ").rstrip(" ")
    input_str = remove_parenthesis_or_brackets(input_str)
    # special case when the ")" is given in a line and ignored by configparser
    input_str = input_str.replace("((", "(")
    # split sub entries
    re_expr = r"\([^\)]*\)"
    frame_grp_str_list = re.findall(re_expr, input_str)
    for frame_grp_str in frame_grp_str_list:
        try:
            frame_grp = FrameGroup.frm_str(frame_grp_str)
        except Exception as e:
            _logger.error(
                f"Unable to create a valid entry from {frame_grp_str}. Error is {e}"
            )
        else:
            result.append(frame_grp)
    return tuple(result)



from __future__ import annotations

import logging
import h5py
from silx.io.utils import h5py_read_dataset

from nxtomo.nxobject.nxdetector import ImageKey
from nxtomomill.models.h52nx._acquisitionstep import AcquisitionStep
from nxtomomill.io.config import TomoHDF5Config

_logger = logging.getLogger(__name__)


class ScanTypeFinder:
    """Util class to find the type of a bliss scan (entry)"""

    def __init__(self, configuration: TomoHDF5Config):
        self._configuration = configuration

    def find(self, entry: h5py.Group):
        if not isinstance(entry, h5py.Group):
            raise ValueError(
                f"Expected path is not related to a h5py.Group ({entry}) when expect to target a bliss entry."
            )
        return (
            self.get_entry_type_from_technique_img_key(entry)
            or self.get_entry_type_from_scan_category(entry)
            or self.get_entry_type_from_title(entry)
        )

    def get_entry_type_from_title(self, entry: h5py.Group):
        """
        try to determine the entry type from the title
        """
        try:
            title = h5py_read_dataset(entry["title"])
        except Exception:
            _logger.error(f"fail to find title for {entry.name}, skip this group")
            return None
        else:
            init_titles = list(self._configuration.init_titles)
            init_titles.extend(self._configuration.zseries_init_titles)
            init_titles.extend(self._configuration.multitomo_init_titles)
            init_titles.extend(self._configuration.back_and_forth_init_titles)

            step_titles = {
                AcquisitionStep.INITIALIZATION: init_titles,
                AcquisitionStep.DARK: self._configuration.dark_titles,
                AcquisitionStep.FLAT: self._configuration.flat_titles,
                AcquisitionStep.PROJECTION: self._configuration.projection_titles,
                AcquisitionStep.ALIGNMENT: self._configuration.alignment_titles,
            }

            for step, titles in step_titles.items():
                for title_start in titles:
                    if title.startswith(title_start):
                        return step
            return None

    def get_entry_type_from_technique_img_key(
        self, entry: h5py.Group
    ) -> AcquisitionStep | None:
        """
        try to determine entry type from the scan/technique sub groups.
        If this is a flat then we expect to have a "flat" group. If this is a set of projection we expect to have a "proj" group.
        For now alignment / return are unfilled
        """
        group_technique = entry.get("technique", dict())
        if "image_key" not in group_technique:
            return None

        image_key = h5py_read_dataset(group_technique["image_key"])
        if image_key is None:
            return None
        else:
            try:
                image_key = ImageKey(image_key)
            except ValueError:
                _logger.error(f"unrecognized image key: '{image_key}'")
                return None
            else:
                connections = {
                    ImageKey.DARK_FIELD: AcquisitionStep.DARK,
                    ImageKey.FLAT_FIELD: AcquisitionStep.FLAT,
                    ImageKey.PROJECTION: AcquisitionStep.PROJECTION,
                    ImageKey.ALIGNMENT: AcquisitionStep.ALIGNMENT,
                }
                return connections.get(image_key, None)

    def get_entry_type_from_scan_category(
        self, entry: h5py.Group
    ) -> AcquisitionStep | None:
        if "scan_category" in entry.get("technique", {}):
            return AcquisitionStep.INITIALIZATION
        else:
            return None



# coding: utf-8

import os
import shutil
import tempfile
from glob import glob

import numpy
import pytest

from tqdm import tqdm

from tomoscan import version
from tomoscan.esrf.mock import MockEDF
from tomoscan.esrf.scan.edfscan import EDFTomoScan
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.esrf.scan.utils import cwd_context, dump_info_file
from tomoscan.validator import is_valid_for_reconstruction
from tomoscan.io import HDF5File, get_swmr_mode

from nxtomomill import converter
from nxtomomill.converter.edf.checks import OUPUT_CHECK
from nxtomomill.models.edf2nx import EDF2nxModel


@pytest.mark.parametrize(
    "duplicate_data, external_link_type",
    ((False, "absolute"), (False, "relative"), (True, "absolute")),
)
@pytest.mark.parametrize("progress", (None, tqdm(desc="conversion from edf")))
@pytest.mark.skipif(
    condition=(version.MINOR < 7 and version.MAJOR == 0),
    reason="dark_n and ref_n from EDFTomoScan where not existing",
)
def test_edf_to_nx_converter(tmp_path, progress, duplicate_data, external_link_type):
    folder = tmp_path

    scan_path = os.path.join(folder, "myscan")
    n_proj = 120
    n_alignment_proj = 5
    dim = 100
    MockEDF(
        scan_path=scan_path,
        n_radio=n_proj,
        n_ini_radio=n_proj,
        n_extra_radio=n_alignment_proj,
        dim=dim,
        dark_n=1,
        flat_n=1,
        distance=2.3,
    )
    scan = EDFTomoScan(scan_path)
    assert scan.dark_n == 1
    output_file = os.path.join(folder, "nexus_file.nx")

    config = EDF2nxModel()
    config.input_folder = scan_path
    config.output_file = output_file
    config.duplicate_data = duplicate_data
    config.external_link_type = external_link_type
    config.x_translation_unit = "cm"
    config.y_translation_unit = "cm"
    config.z_translation_unit = "m"
    config.sample_detector_distance_unit = "cm"
    nx_file, nx_entry = converter.from_edf_to_nx(
        configuration=config,
        progress=progress,
    )

    hdf5_scan = NXtomoScan(scan=nx_file, entry=nx_entry)
    assert len(hdf5_scan.projections) == n_proj
    assert len(hdf5_scan.alignment_projections) == n_alignment_proj
    assert hdf5_scan.dim_1 == dim
    assert hdf5_scan.dim_2 == dim
    assert is_valid_for_reconstruction(hdf5_scan)
    # insure links are valid in case of external
    # for relative links it looks we still need to be on the file working directory to solve links
    # see policy details here: https://docs.hdfgroup.org/hdf5/v1_12/group___h5_l.html#title5
    # this will work with the tomotools suite as tomoscan will automatically create a working dircetory context when reading the file.
    # but this can be an issue when sharing data or reading data outside the tomotools suite
    with cwd_context(output_file):
        with HDF5File(output_file, mode="r", swmr=get_swmr_mode()) as h5f:
            detector_data_path = f"/{nx_entry}/instrument/detector/data"
            assert (
                f"/{nx_entry}/instrument/detector/data" in h5f
            ), f"/{nx_entry}/instrument/detector/data  doesn't exists"
            assert h5f[detector_data_path].is_virtual == (not duplicate_data)
            assert (
                h5f[detector_data_path][()].min() != h5f[detector_data_path][()].max()
            )
            numpy.testing.assert_almost_equal(
                h5f[f"/{nx_entry}/sample/z_translation"][0], 0.0
            )
            numpy.testing.assert_almost_equal(
                h5f[f"/{nx_entry}/sample/x_translation"][0], 1.0 / 100.0
            )
            numpy.testing.assert_almost_equal(
                h5f[f"/{nx_entry}/sample/y_translation"][0], 2.0
            )

    # x, y and z are set manually on MockEDF.
    n_frames = 120 + 5 + 1 + 1
    numpy.testing.assert_array_almost_equal(
        hdf5_scan.x_translation,
        numpy.array([1.0 / 100.0] * n_frames),
    )
    numpy.testing.assert_array_almost_equal(
        hdf5_scan.y_translation, numpy.array([2.0] * n_frames)
    )
    numpy.testing.assert_almost_equal(hdf5_scan.distance, 2.3 / 100.0)


@pytest.mark.parametrize("scan_range", (-180, 180, 360))
@pytest.mark.parametrize("endpoint", (True, False))
@pytest.mark.parametrize("revert", (True, False))
@pytest.mark.parametrize("force_angles", (True, False))
def test_rotation_angle_infos(scan_range, endpoint, revert, force_angles):
    """test conversion fits EDF2nxModel parameters regarding the rotation angle calculation options"""
    with tempfile.TemporaryDirectory() as data_folder:
        scan_path = os.path.join(data_folder, "myscan")
        n_proj = 12
        n_alignment_proj = 5
        dim = 4

        MockEDF(
            scan_path=scan_path,
            n_radio=n_proj,
            n_ini_radio=n_proj,
            n_extra_radio=n_alignment_proj,
            dim=dim,
            dark_n=1,
            ref_n=1,
            scan_range=scan_range,
            rotation_angle_endpoint=endpoint,
        )

        output_file = os.path.join(data_folder, "nexus_file.nx")

        config = EDF2nxModel()
        config.input_folder = scan_path
        config.output_file = output_file
        config.force_angle_calculation = force_angles
        config.angle_calculation_endpoint = endpoint
        config.angle_calculation_rev_neg_scan_range = revert

        nx_file, nx_entry = converter.from_edf_to_nx(
            configuration=config,
        )
        hdf5_scan = NXtomoScan(scan=nx_file, entry=nx_entry)

        # compute expected rotation angles
        raw_rotation_angles = numpy.asarray(hdf5_scan.rotation_angle)
        converted_rotation_angles = raw_rotation_angles[
            hdf5_scan.image_key_control == 0
        ]

        if force_angles and revert and scan_range < 0 and not endpoint:
            expected_angles = numpy.linspace(0, scan_range, n_proj, endpoint=endpoint)
        else:
            expected_angles = numpy.linspace(
                min(0, scan_range), max(0, scan_range), n_proj, endpoint=endpoint
            )
            revert_angles_in_nx = force_angles and revert and (scan_range < 0)
            if revert_angles_in_nx:
                expected_angles = expected_angles[::-1]

        numpy.testing.assert_almost_equal(
            converted_rotation_angles, expected_angles, decimal=3
        )
        # test alignment projection and flat are contained in the projections range
        dark_angles = raw_rotation_angles[hdf5_scan.image_key_control == 2]
        for angle in dark_angles:
            assert angle == 0 or min(converted_rotation_angles) <= angle <= max(
                converted_rotation_angles
            )

        flat_angles = raw_rotation_angles[hdf5_scan.image_key_control == 1]

        for angle in flat_angles:
            assert angle == 0 or min(converted_rotation_angles) <= angle <= max(
                converted_rotation_angles
            )

        alignment_angles = raw_rotation_angles[hdf5_scan.image_key_control == -1]
        for angle in alignment_angles:
            assert angle == 0 or min(converted_rotation_angles) <= angle <= max(
                converted_rotation_angles
            )


def test_rot_angle_key_does_not_exists():
    """test conversion fits EDF2nxModel parameters regarding the rotation angle calculation options"""
    with tempfile.TemporaryDirectory() as data_folder:
        scan_path = os.path.join(data_folder, "myscan")
        n_proj = 12
        n_alignment_proj = 5
        dim = 4

        MockEDF(
            scan_path=scan_path,
            n_radio=n_proj,
            n_ini_radio=n_proj,
            n_extra_radio=n_alignment_proj,
            dim=dim,
            dark_n=1,
            ref_n=1,
            scan_range=180,
        )

        output_file = os.path.join(data_folder, "nexus_file.nx")

        config = EDF2nxModel()
        config.input_folder = scan_path
        config.output_file = output_file
        config.force_angle_calculation = False
        config.rotation_angle_keys = tuple()

        nx_file, nx_entry = converter.from_edf_to_nx(
            configuration=config,
        )
        hdf5_scan = NXtomoScan(scan=nx_file, entry=nx_entry)

        # compute expected rotation angles
        raw_rotation_angles = numpy.asarray(hdf5_scan.rotation_angle)
        converted_rotation_angles = raw_rotation_angles[
            hdf5_scan.image_key_control == 0
        ]

        expected_angles = numpy.linspace(
            0, 180, n_proj, endpoint=config.angle_calculation_endpoint
        )
        numpy.testing.assert_almost_equal(
            converted_rotation_angles, expected_angles, decimal=3
        )


def test_different_info_file():
    """insure providing a different spec info file will be taken into account"""
    with tempfile.TemporaryDirectory() as data_folder:
        scan_path = os.path.join(data_folder, "myscan")
        n_proj = 12
        n_alignment_proj = 0
        dim = 4
        flat_n = 1
        dark_n = 1

        original_scan_range = 180
        original_energy = 2.3
        original_pixel_size = 0.03
        original_distance = 0.36

        new_scan_range = -180
        new_energy = 6.6
        new_pixel_size = 0.002
        new_distance = 0.458
        new_n_proj = n_proj - 2

        other_info_file = os.path.join(data_folder, "new_info_file.info")
        dump_info_file(
            file_path=other_info_file,
            tomo_n=new_n_proj,
            scan_range=new_scan_range,
            flat_n=flat_n,
            flat_on=new_n_proj,
            dark_n=dark_n,
            dim_1=dim,
            dim_2=dim,
            col_beg=0,
            col_end=dim,
            row_beg=0,
            row_end=dim,
            pixel_size=new_pixel_size,
            distance=new_distance,
            energy=new_energy,
        )
        assert os.path.exists(other_info_file)

        MockEDF(
            scan_path=scan_path,
            n_radio=n_proj,
            n_ini_radio=n_proj,
            n_extra_radio=n_alignment_proj,
            dim=dim,
            dark_n=dark_n,
            flat_n=flat_n,
            scan_range=original_scan_range,
            energy=original_energy,
            pixel_size=original_pixel_size,
            distance=original_distance,
        )

        output_file = os.path.join(data_folder, "nexus_file.nx")
        config = EDF2nxModel()
        config.input_folder = scan_path
        config.output_file = output_file
        config.force_angle_calculation = True
        config.pixel_size_unit = "m"
        config.distance_unit = "m"
        config.energy_unit = "keV"

        # test step 1: check scan info is correctly read from the original parameters
        nx_file, nx_entry = converter.from_edf_to_nx(
            configuration=config,
        )
        hdf5_scan_original_info_file = NXtomoScan(scan=nx_file, entry=nx_entry)
        assert numpy.isclose(hdf5_scan_original_info_file.energy, original_energy)
        assert hdf5_scan_original_info_file.scan_range == original_scan_range
        assert hdf5_scan_original_info_file.pixel_size == original_pixel_size
        assert hdf5_scan_original_info_file.distance == original_distance
        assert len(hdf5_scan_original_info_file.projections) == n_proj

        # test step 2: check scan info is correctly read from new parameters
        config.dataset_info_file = other_info_file
        config.overwrite = True
        nx_file, nx_entry = converter.from_edf_to_nx(
            configuration=config,
        )
        hdf5_scan_new_info_file = NXtomoScan(scan=nx_file, entry=nx_entry)
        assert numpy.isclose(hdf5_scan_new_info_file.energy, new_energy)
        # for now NXtomoScan expect range to be 180 or 360. There is no such -180 as in EDF
        assert abs(hdf5_scan_new_info_file.scan_range) == abs(new_scan_range)
        assert hdf5_scan_new_info_file.pixel_size == new_pixel_size
        assert hdf5_scan_new_info_file.distance == new_distance
        assert len(hdf5_scan_new_info_file.projections) == new_n_proj


def test_different_dataset_basename():
    """test conversion succeed if we provide a dataset with a different basename"""
    with tempfile.TemporaryDirectory() as data_folder:
        original_scan_path = os.path.join(data_folder, "myscan")
        new_scan_path = os.path.join(data_folder, "myscan_435")

        n_proj = 12
        n_alignment_proj = 5
        dim = 4
        energy = 12.35
        n_darks = 2
        n_flats = 1

        MockEDF(
            scan_path=original_scan_path,
            n_radio=n_proj,
            n_ini_radio=n_proj,
            n_extra_radio=n_alignment_proj,
            dim=dim,
            dark_n=n_darks,
            flat_n=n_flats,
            energy=energy,
        )

        shutil.move(
            original_scan_path,
            new_scan_path,
        )

        output_file = os.path.join(data_folder, "nexus_file.nx")

        config = EDF2nxModel()
        config.input_folder = new_scan_path
        config.output_file = output_file
        config.dataset_basename = "myscan"
        nx_file, nx_entry = converter.from_edf_to_nx(
            configuration=config,
        )
        hdf5_scan = NXtomoScan(scan=nx_file, entry=nx_entry)
        assert len(hdf5_scan.projections) == n_proj
        assert len(hdf5_scan.alignment_projections) == n_alignment_proj
        assert len(hdf5_scan.darks) == n_darks
        assert len(hdf5_scan.flats) == n_flats
        assert hdf5_scan.energy == energy


def test_delete_edf_input_files():
    """
    test `delete_edf_source_files` option
    """

    def get_n_edf_file(path):
        return len(glob(os.path.join(path, "*.edf")))

    with tempfile.TemporaryDirectory() as data_folder:
        input_scan_path = os.path.join(data_folder, "input_scan")

        MockEDF(
            scan_path=input_scan_path,
            n_radio=10,
            n_ini_radio=10,
            n_extra_radio=2,
            dim=25,
            dark_n=2,
            flat_n=2,
        )
        assert get_n_edf_file(input_scan_path) == 16
        info_file = os.path.join(input_scan_path, "input_scan.info")
        assert os.path.exists(info_file)

        config = EDF2nxModel()

        nx_tomo_path = os.path.join(data_folder, "nxtomo.nx")
        config.input_folder = input_scan_path
        config.output_file = nx_tomo_path
        config.delete_edf_source_files = True
        config.duplicate_data = False
        # make sure if we ask for no data duplication and removing source files we raise an error (incoherent settings)
        with pytest.raises(ValueError):
            converter.from_edf_to_nx(
                configuration=config,
            )

        config.duplicate_data = True
        nx_file, _ = converter.from_edf_to_nx(
            configuration=config,
        )
        # check conversion went well
        assert os.path.exists(nx_file)
        # make sure there is no edf left
        assert get_n_edf_file(input_scan_path) == 0
        # make sure the .info file is still there to
        assert os.path.exists(info_file)


def test_output_checks():
    """
    test some conversion calling some check at the end of the conversion
    """
    with tempfile.TemporaryDirectory() as data_folder:
        input_scan_path = os.path.join(data_folder, "input_scan")

        MockEDF(
            scan_path=input_scan_path,
            n_radio=10,
            n_ini_radio=10,
            n_extra_radio=2,
            dim=25,
            dark_n=2,
            flat_n=2,
        )

        config = EDF2nxModel()
        nx_tomo_path = os.path.join(data_folder, "nxtomo.nx")
        config.input_folder = input_scan_path
        config.output_file = nx_tomo_path
        config._output_checks = (OUPUT_CHECK.COMPARE_VOLUME,)
        converter.from_edf_to_nx(
            configuration=config,
        )



from nxtomomill.utils.io import deprecated_warning
from nxtomomill.tests import *  # noqa F401

deprecated_warning(
    type_="Module",
    name="nxtomomill.test",
    reason="renamed",
    replacement="nxtomomill.tests",
    since_version=1.1,
)



from __future__ import annotations

from silx.io.url import DataUrl
from silx.io.utils import open as open_hdf5

from nxtomomill.io.config import TomoHDF5Config
from nxtomomill.models.h52nx._acquisitionstep import AcquisitionStep
from nxtomomill.utils.io import deprecated

from ._scan_type_finder import ScanTypeFinder


@deprecated(reason="renamed", replacement="get_bliss_scan_type", since_version="2.0")
def get_entry_type(*args, **kwargs):
    return get_bliss_scan_type(*args, **kwargs)


def get_bliss_scan_type(
    url: DataUrl, configuration: TomoHDF5Config
) -> AcquisitionStep | None:
    """
    :param url: bliss scan url to type
    :return: return the step of the acquisition or None if cannot find it.
    """
    if not isinstance(url, DataUrl):
        raise TypeError(f"DataUrl is expected. Not {type(url)}")
    if url.data_slice() is not None:
        raise ValueError(
            "url is expected to reference a Bliss scan entry (no slices supported)"
        )
    scan_type_finder = ScanTypeFinder(configuration=configuration)

    with open_hdf5(url.file_path()) as h5f:
        if url.data_path() not in h5f:
            raise ValueError(f"Provided path does not exist: {url}")
        entry = h5f[url.data_path()]
        return scan_type_finder.find(entry)



from __future__ import annotations

from nxtomomill.models.utils import convert_str_to_tuple, filter_str_def

from ..base.general_section import GeneralSection as _GeneralSectionBase

from pydantic import Field, field_validator, ConfigDict


class GeneralSection(_GeneralSectionBase):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    input_folder: str | None = Field(
        default=None,
        description="Path to the folder containing the raw data folder and the fluofit/ subfolder. if not provided from the configuration file must be provided from the command line",
    )
    dimension: int = Field(
        default=3,
        description="Dimension of the experiment. 2 for 2D XRFCT, 3 for 3D XRFCT. Default is 3.",
    )
    detector_names: tuple[str, ...] = Field(
        default=tuple(),
        description="Define a list of (real or virtual) detector names used for the exp (space separated values - no comma). E.g. 'falcon xmap'. If not specified, all detectors are processed.",
    )
    dataset_basename: str | None = Field(
        default=None,
        description="In 2D, the exact full name of the folder. In 3D, the folder name prefix (the program will search for folders named <prefix>_XXX where XXX is a nmber.) If not provided will take the name of input_folder",
    )
    dataset_info_file: str | None = Field(
        default=None,
        description="Path to .info file containing dataset information (Energy, ScanRange, TOMO_N...). If not will deduce it from dataset_basename",
    )
    patterns_to_ignores: tuple[str, ...] = Field(
        default=("_slice_",),
        description="Some file pattern leading to ignoring the file. Like reconstructed slice files.",
    )
    duplicate_data: bool = Field(
        default=True,
        description="If False then will create embed all the data into a single file avoiding external link to other file. If True then the decetor data will point to original tif files. In this case you must be carreful to keep relative paths valid. Warning: to read external dataset you nust be at the hdf5 file working directory. See external link resolution details: https://support.hdfgroup.org/documentation/hdf5/latest/group___h5_l.html#title5",
    )

    @field_validator(
        "detector_names",
        "patterns_to_ignores",
        mode="before",
    )
    @classmethod
    def cast_to_tuple(cls, value: str | None) -> tuple[str]:
        return convert_str_to_tuple(value) or ()

    @field_validator(
        "input_folder",
        "dataset_basename",
        "dataset_info_file",
        mode="plain",
    )
    @classmethod
    def cast_to_str(cls, value: str | None) -> tuple | None:
        return filter_str_def(value)



# coding: utf-8

from __future__ import annotations

import logging
import os

import numpy
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.esrf.scan.utils import cwd_context
from tomoscan.framereducer.method import ReduceMethod
from silx.io.utils import open as open_hdf5

from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey

from ..utils.utils import strip_extension

logging.basicConfig(level=logging.INFO)

_logger = logging.getLogger(__name__)

__all__ = ["flat_reducer", "extract_darks_flats"]


def extract_darks_flats(
    dataset_file_name: str,
    entry_name: str,
    save_intermediated: bool = False,
    target_filename: str | None = None,
    target_entry_name: str | None = None,
    method: str = "median",
    reuse_intermediated: bool = False,
    use_projections_for_flats: bool = False,
    dark_default_value=None,
):
    dataset_file_name = os.path.abspath(dataset_file_name)
    target_entry_name = target_entry_name if target_entry_name else entry_name

    dirname = os.path.dirname(dataset_file_name)

    basename = os.path.basename(dataset_file_name)

    if not dirname:
        dirname = "./"

    if target_filename is not None:
        target_filename = os.path.abspath(target_filename)

    with cwd_context(dirname):
        if reuse_intermediated:
            scan = NXtomoScan(target_filename, target_entry_name)
            reduced_flats, metadata_flats = scan.load_reduced_flats(return_info=True)
            reduced_darks, metadata_darks = scan.load_reduced_darks(return_info=True)
        else:
            nxt = NXtomo()
            nxt.load(basename, data_path=entry_name)
            if use_projections_for_flats:
                where_proj = [k.value == 0 for k in nxt.instrument.detector.image_key]
                where_flat = [k.value == 1 for k in nxt.instrument.detector.image_key]

                nxt.instrument.detector.image_key_control[where_proj] = (
                    ImageKey.FLAT_FIELD
                )
                nxt.instrument.detector.image_key_control[where_flat] = ImageKey.INVALID

                file_path = f"{basename}_edited_keys_scan.nx"
                if os.path.isfile(file_path):
                    os.remove(file_path)
                nxt.save(file_path, entry_name)

                scan = NXtomoScan(file_path, entry_name)
                reduced_flats, metadata_flats = scan.compute_reduced_flats(
                    method, return_info=True
                )
                reduced_darks, metadata_darks = scan.compute_reduced_darks(
                    return_info=True
                )
                if len(reduced_darks) == 0:
                    assert len(reduced_flats), " We expect to find at least  some flats"
                    dim_2, dim_1 = reduced_flats[list(reduced_flats.keys())[0]].shape
                    _logger.warning(
                        f" patching with a default dark of size {dim_1} for horizontal , {dim_2} for vertical and default value {dark_default_value}"
                    )
                    assert (
                        dark_default_value is not None
                    ) > 0, f"No raw darks found in the dataset {scan} and 'dark_default_value' not provided. Unable to get any reduced darks."

                    reduced_darks[0] = numpy.full(
                        (dim_2, dim_1), dark_default_value, dtype="f"
                    )
                    metadata_darks = metadata_flats
            else:
                scan = NXtomoScan(basename, entry_name)
                reduced_flats, metadata_flats = scan.compute_reduced_flats(
                    method, return_info=True
                )
                reduced_darks, metadata_darks = scan.compute_reduced_darks(
                    return_info=True
                )
                reduced_flats, metadata_flats = scan.compute_reduced_flats(
                    method, return_info=True
                )

        if save_intermediated:
            scan = NXtomoScan(target_filename, target_entry_name)
            scan.save_reduced_flats(
                reduced_flats, flats_infos=metadata_flats, overwrite=True
            )
            scan.save_reduced_darks(
                reduced_darks, darks_infos=metadata_darks, overwrite=True
            )

    return_dict = {
        "flat": {"images": reduced_flats, "meta": metadata_flats},
        "dark": {"images": reduced_darks, "meta": metadata_darks},
    }

    return __RefsDarks(return_dict, entry_name), return_dict


class __RefsDarks:
    def __init__(self, dict_or_file_name, entry_name):
        self.dict_or_file_name = dict_or_file_name
        self.entry_name = entry_name
        self.flat_image, self.flat_current = self._take_image_and_meta("flat")
        self.dark_image, self.dark_current = self._take_image_and_meta("dark")

    def _take_image_and_meta(self, what) -> tuple:
        """
        :return: a tuple as (image, current:float|None)
        """
        if isinstance(self.dict_or_file_name, dict):
            group = self.dict_or_file_name[what]  # [self.entry_name]
            image = None
            for key in group["images"]:
                if (
                    isinstance(key, int)
                    or (isinstance(key, str) and key.isnumeric())
                    or (numpy.isdtype(numpy.dtype(key), "integral"))
                ):
                    if image is None:
                        image = group["images"][key]
                    else:
                        _logger.warning("More than one image found.")
            if len(group["meta"].machine_current) > 0:
                current = group["meta"].machine_current[0]
            else:
                current = None
        else:
            file_name_tmp = f"{strip_extension(self.dict_or_file_name)}_{what}.h5"
            with open_hdf5(file_name_tmp) as f:
                group = f[self.entry_name]
                group = f[what]
                image = None
                current = group["machine_current"][()][0]
                for key in group:
                    if (
                        isinstance(key, int)
                        or (isinstance(key, str) and key.isnumeric())
                        or (numpy.isdtype(numpy.dtype(key), "integral"))
                    ):
                        if image is None:
                            image = group[key][()]
                        else:
                            raise ValueError(
                                f"More than one image found in {file_name_tmp}"
                            )

        return image, current


def flat_reducer(
    scan_filename: str,
    ref_start_filename: str,
    ref_end_filename: str,
    mixing_factor: float,
    entry_name: str = "entry0000",
    median_or_mean: str = ReduceMethod.MEAN.value,
    save_intermediated: bool = False,
    reuse_intermediated: bool = False,
    overwrite: bool = True,
    dark_default_value=300,
):
    """
    this method extracts first a flatfield and dark from  two  reference scans. After flats and darks extraction, an interpolation is done
    according to the mixing_factor parameter. The obtained flats and dark are then saved associating them for a given target scan_filename

    :param scan_filename: The target scan. A nexus filename for which we want to create reduced scan from the scans
        given by ref_start and ref_end parameters ( a scan at the beginning, another at the end)
    :param ref_start_filename: The scan with projections to be used as reference for the beginning of the measures.
    :param ref_end_filename: The scan with projections to be used as reference at the end  of the measures.
    :param mixing_factor: The mixing factor giving the averaged flats as
        (ref_start-darkB+darkS)*(1-mixing_factor)+(ref_end-darkE+darkS)*mixing_factor
    :param entry_name: The entry name, it defaults to entry0000
    :param median_or_mean: Either "mean" or "median". Default is "mean"
    :param save_intermediated: Save intermediated flats and darks corresponding to extremal
        reference scans (ref_start_filename, refa_filename) for later usage. Defaults to False
    :param use_intermediated: Save  intermediated flats and darks and if already presente reuse them for mixing
    :param overwrite: enforce overwriting of the reduced flats/darks
    """

    if reuse_intermediated:
        required_files = [
            f"{strip_extension(ref_start_filename, _logger)}_darks.hdf5",
            f"{strip_extension(ref_start_filename, _logger)}_flats.hdf5",
            f"{strip_extension(ref_end_filename, _logger)}_darks.hdf5",
            f"{strip_extension(ref_end_filename, _logger)}_flats.hdf5",
        ]
        intermediated_are_reusable = True
        for fn in required_files:
            if not os.path.exists(fn):
                intermediated_are_reusable = False
    else:
        intermediated_are_reusable = False

    # saving the intermediae if enforced if there is a plan to use them
    # and they are not available yet
    save_intermediated = save_intermediated or (
        reuse_intermediated and not intermediated_are_reusable
    )

    if median_or_mean not in [ReduceMethod.MEAN.value, ReduceMethod.MEDIAN.value]:
        message = f""" the "median_or_mean" parameter must be one of {[ReduceMethod.MEAN.value, ReduceMethod.MEDIAN.value]}.
        It was {median_or_mean}
        """
        raise ValueError(message)

    fd_start, fd_start_as_dict = extract_darks_flats(
        ref_start_filename,
        entry_name,
        target_filename=ref_start_filename,
        save_intermediated=save_intermediated,
        method=median_or_mean,
        reuse_intermediated=intermediated_are_reusable,
        use_projections_for_flats=True,
        dark_default_value=dark_default_value,
    )

    fd_end, _ = extract_darks_flats(
        ref_end_filename,
        entry_name,
        target_filename=ref_end_filename,
        save_intermediated=save_intermediated,
        method=median_or_mean,
        reuse_intermediated=intermediated_are_reusable,
        use_projections_for_flats=True,
        dark_default_value=dark_default_value,
    )
    fd_sample, fd_as_dict = extract_darks_flats(
        scan_filename,
        entry_name,
        method=median_or_mean,
        use_projections_for_flats=False,
    )
    reduced_infos = fd_as_dict["flat"]["meta"]

    scan = NXtomoScan(scan_filename, entry_name)
    current = fd_sample.flat_current
    if current is None:
        # handle the case the fd_sample does not contains any flat frames. In this case get the first
        # current we find from the NXtomo
        currents = scan.machine_current
        if currents is not None and len(currents) > 0:
            current = currents[0]  # pylint: disable=E1136

    if current is None:
        raise ValueError(
            f"Unable to find any machine current from {scan_filename}. Unable to compute reduced darks and flats"
        )

    # compute reduced flats and dark
    flat0 = (
        fd_start.flat_image - fd_start.dark_image
    ) * current / fd_start.flat_current + fd_start.dark_image
    flat1 = (
        fd_end.flat_image - fd_start.dark_image
    ) * current / fd_end.flat_current + fd_start.dark_image

    flat = (1 - mixing_factor) * flat0 + mixing_factor * flat1

    reduced_flats = {0: flat}

    # save reduced flats and dark
    reduced_infos.machine_current = numpy.array([current])
    reduced_infos.count_time = reduced_infos.count_time[:1]
    if current != reduced_infos.machine_current[0]:
        raise RuntimeError(
            " Coherence check failed. Total non sense: the code is broken."
        )

    scan.save_reduced_flats(
        reduced_flats, flats_infos=reduced_infos, overwrite=overwrite
    )

    reduced_darks = fd_start_as_dict["dark"]["images"]
    reduced_infos = fd_start_as_dict["dark"]["meta"]
    scan.save_reduced_darks(
        reduced_darks, darks_infos=reduced_infos, overwrite=overwrite
    )



# coding: utf-8

import os
import numpy
import pytest
import h5py

from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill.io.config import TomoHDF5Config
from nxtomomill.models.h52nx._acquisitionstep import AcquisitionStep
from nxtomomill.converter.hdf5.acquisition.utils import (
    deduce_machine_current,
    split_timestamps,
    get_bliss_scan_type,
    group_series,
)
from nxtomomill.converter.hdf5.acquisition.zseriesacquisition import (
    ZSeriesBaseAcquisition,
)
from nxtomomill.utils.utils import str_datetime_to_numpy_datetime64
from nxtomomill.converter.hdf5.acquisitionConstructor import (
    is_multitomo_sequence,
    is_back_and_forth_sequence,
)

from nxtomomill.converter.hdf5.acquisition.tests.bliss_tomo_datasets import (  # noqa F401
    holotomo2_simu_dataset,
)
from nxtomomill.converter.hdf5.acquisition.utils._scan_type_finder import ScanTypeFinder


def test_deduce_machine_current():
    """
    Test `deduce_current` function. Base function to compute current for each frame according to it's timestamp
    """

    current_datetimes = {
        "2022-01-15T21:07:58.360095+02:00": 1.1,
        "2022-04-15T21:07:58.360095+02:00": 121.1,
        "2022-04-15T21:09:58.360095+02:00": 123.3,
        "2022-04-15T21:11:58.360095+02:00": 523.3,
        "2022-12-15T21:07:58.360095+02:00": 1000.3,
    }
    with pytest.raises(ValueError):
        deduce_machine_current(tuple(), {})
    with pytest.raises(TypeError):
        deduce_machine_current(12, 2)
    with pytest.raises(TypeError):
        deduce_machine_current(
            [
                12,
            ],
            2,
        )
    with pytest.raises(TypeError):
        deduce_machine_current(
            [
                2,
            ],
            current_datetimes,
        )

    converted_currents = {}
    for elec_cur_datetime_str, elect_cur in current_datetimes.items():
        datetime_as_datetime = str_datetime_to_numpy_datetime64(elec_cur_datetime_str)
        converted_currents[datetime_as_datetime] = elect_cur

    # check exacts values, left and right bounds
    assert deduce_machine_current(
        (str_datetime_to_numpy_datetime64("2022-01-15T21:07:58.360095+02:00"),),
        converted_currents,
    ) == (1.1,)
    assert deduce_machine_current(
        (str_datetime_to_numpy_datetime64("2022-01-14T21:07:58.360095+02:00"),),
        converted_currents,
    ) == (1.1,)
    assert deduce_machine_current(
        (str_datetime_to_numpy_datetime64("2022-12-15T21:07:58.360095+02:00"),),
        converted_currents,
    ) == (1000.3,)
    assert deduce_machine_current(
        (str_datetime_to_numpy_datetime64("2022-12-16T21:07:58.360095+02:00"),),
        converted_currents,
    ) == (1000.3,)
    # check interpolated values
    numpy.testing.assert_almost_equal(
        deduce_machine_current(
            (str_datetime_to_numpy_datetime64("2022-04-15T21:08:58.360095+02:00"),),
            converted_currents,
        )[0],
        (122.2,),
    )
    numpy.testing.assert_almost_equal(
        deduce_machine_current(
            (str_datetime_to_numpy_datetime64("2022-04-15T21:10:28.360095+02:00"),),
            converted_currents,
        )[0],
        (223.3,),
    )

    # test several call and insure keep order
    numpy.testing.assert_almost_equal(
        deduce_machine_current(
            (
                str_datetime_to_numpy_datetime64("2022-01-15T21:07:58.360095+02:00"),
                str_datetime_to_numpy_datetime64("2022-04-15T21:10:28.360095+02:00"),
                str_datetime_to_numpy_datetime64("2022-04-15T21:08:58.360095+02:00"),
            ),
            converted_currents,
        ),
        (1.1, 223.3, 122.2),
    )


@pytest.mark.parametrize("n_part", (1, 3, 4, 7, 9, 12))
@pytest.mark.parametrize(
    "array_to_test",
    (numpy.arange(-15, 23, dtype=numpy.float32), numpy.ones(57, dtype=numpy.uint8)),
)
def test_split_timestamps(n_part, array_to_test):
    split_arrays = tuple(
        split_timestamps(
            my_array=array_to_test,
            n_part=n_part,
        )
    )
    assert len(split_arrays) == n_part
    assert split_arrays[0].dtype == array_to_test.dtype
    numpy.testing.assert_array_equal(
        numpy.concatenate([numpy.array(part) for part in split_arrays]),
        numpy.array(array_to_test),
    )


def test_get_entry_type(tmp_path):
    """test the get_entry_type function"""
    test_folder = tmp_path / "test_get_entry_type"
    test_folder.mkdir()
    test_file = os.path.join(test_folder, "test.hdf5")

    default_configuration = TomoHDF5Config()

    with HDF5File(test_file, mode="w") as h5f:
        h5f["group1/title"] = "darks"
        url_case1_darks = DataUrl(file_path=test_file, data_path="group1")

        h5f["group2/title"] = "flats"
        url_case1_flats = DataUrl(file_path=test_file, data_path="group2")

        h5f["group3/title"] = "darks"
        h5f["group3/technique/image_key"] = ImageKey.PROJECTION.value
        url_case2_proj = DataUrl(file_path=test_file, data_path="group3")

        h5f["group4/title"] = "darks"
        h5f["group4/technique/image_key"] = ImageKey.ALIGNMENT.value
        url_case2_alignment = DataUrl(file_path=test_file, data_path="group4")

        h5f["group5/technique/image_key"] = ImageKey.INVALID.value
        url_case3_invalid = DataUrl(file_path=test_file, data_path="group5")
        h5f["group6/technique/image_key"] = ImageKey.PROJECTION.value
        url_case3_proj = DataUrl(file_path=test_file, data_path="group6")

    # make sure if title is provided this is still working
    assert (
        get_bliss_scan_type(url_case1_darks, default_configuration)
        == AcquisitionStep.DARK
    )
    assert (
        get_bliss_scan_type(url_case1_flats, default_configuration)
        == AcquisitionStep.FLAT
    )

    # make sure the 'technique/image_key' get priority other the 'title'
    assert (
        get_bliss_scan_type(url_case2_proj, default_configuration)
        == AcquisitionStep.PROJECTION
    )
    assert (
        get_bliss_scan_type(url_case2_alignment, default_configuration)
        == AcquisitionStep.ALIGNMENT
    )

    # make sure if title is no more present then 'image_key' will be picked anyway
    assert get_bliss_scan_type(url_case3_invalid, default_configuration) is None
    assert (
        get_bliss_scan_type(url_case3_proj, default_configuration)
        == AcquisitionStep.PROJECTION
    )


def test_group_z_series(tmp_path):
    input_file = os.path.join(tmp_path, "test_z_series.h5")
    z_acquisitions = []
    n_series = 3
    n_scan_per_series = 4
    with h5py.File(input_file, mode="w") as h5f:
        i_scan = 0
        for i_series in range(n_series):
            for _ in range(n_scan_per_series):
                h5f[f"{i_scan}.1/sample/name"] = (
                    f"my_serie_{str(i_series).zfill(4)}_{str(i_scan).zfill(4)}"
                )
                z_acquisitions.append(
                    ZSeriesBaseAcquisition(
                        root_url=DataUrl(
                            file_path=input_file,
                            data_path=f"{i_scan}.1",
                            scheme="silx",
                        ),
                        configuration={},
                        detector_sel_callback=None,
                        start_index=i_scan,
                    )
                )
                i_scan += 1

    series_grouped = []
    for acquisition in z_acquisitions:
        series_grouped = group_series(
            acquisition=acquisition, list_of_series=series_grouped
        )

    assert len(series_grouped) == n_series
    for i in range(n_series):
        assert series_grouped[i] == list(
            [
                z_acquisitions[i * n_scan_per_series + i_scan]
                for i_scan in range(n_scan_per_series)
            ]
        )


@pytest.mark.parametrize(
    "bliss_scan_path, scan_type",
    {
        "1.1": AcquisitionStep.INITIALIZATION,  # for now this is detected as an 'initialization' but in the absolute it should be a sequence of sequence.
        "2.1": AcquisitionStep.INITIALIZATION,
        "3.1": AcquisitionStep.PROJECTION,
    }.items(),
)
def test_holotomo2_simu_and_acquisition_type(
    holotomo2_simu_dataset, bliss_scan_path, scan_type  # noqa F811
):
    file_path = os.path.join(
        holotomo2_simu_dataset, "sample", "sample_holotomo2", "sample_holotomo2.h5"
    )
    configuration = TomoHDF5Config()
    scan_type_finder = ScanTypeFinder(configuration)
    with h5py.File(file_path, mode="r") as h5f:
        entry = h5f[bliss_scan_path]
        ScanTypeFinder
        # dummy test of 'is_multitomo_sequence'
        assert (
            is_multitomo_sequence(
                entry,
                configuration=configuration,
                fallback_on_title=False,
            )
            is False
        )
        assert (
            is_back_and_forth_sequence(
                entry,
                configuration=configuration,
                fallback_on_title=False,
            )
            is False
        )
        assert scan_type_finder.find(entry=entry) is scan_type



# coding: utf-8

from __future__ import annotations

from pydantic import BaseModel, ConfigDict

from .general_section import GeneralSection
from ..base.source_section import SourceSection
from ..base.instrument_section import InstrumentSection
from ..base.NestedModelBase import NestedModelBase

__all__ = [
    "BlissFluo2nxModel",
]


class BlissFluo2nxModel(
    BaseModel,
    NestedModelBase,
):
    model_config = ConfigDict(str_to_upper=True, validate_assignment=True)

    general_section: GeneralSection = GeneralSection()
    source_section: SourceSection = SourceSection()
    instrument_section: InstrumentSection = InstrumentSection()

    def model_dump(self, *args, **kwargs) -> dict:
        unordered_result = super().model_dump(*args, **kwargs)
        orderdered_result = {
            "ewoksfluo_filename": unordered_result.pop("ewoksfluo_filename"),
            "output_file": unordered_result.pop("output_file"),
            "detector_names": unordered_result.pop("detector_names"),
            "dimension": unordered_result.pop("dimension"),
        }
        orderdered_result.update(unordered_result)
        return orderdered_result



from __future__ import annotations

from enum import Enum

from pydantic import (
    BaseModel,
    field_validator,
    field_serializer,
    ConfigDict,
    Field,
)

from silx.io.url import DataUrl

from ._acquisitionstep import AcquisitionStep

from nxtomomill.models.utils import filter_str_def
from nxtomomill.models.utils import (
    remove_parenthesis_or_brackets,
    convert_str_to_bool,
)


class FrameGroup(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True, validate_assignment=True, validate_by_name=True
    )

    url: DataUrl | None = None
    frame_type: AcquisitionStep = AcquisitionStep.PROJECTION
    copy_data: bool | None = Field(
        default=None,
        description="Should the frame dataset be copied or not. If not set will fallback on the 'frame_type_section.default_data_copy'",
    )

    class Info(Enum):
        URL_ENTRY = "entry"
        FRAME_TYPE = "frame_type"
        COPY = "copy"

    @field_validator(
        "frame_type",
        mode="plain",
    )
    @classmethod
    def cast_to_AcquisitionStep(cls, value: str | AcquisitionStep) -> AcquisitionStep:
        return AcquisitionStep.from_value(value)

    @field_validator(
        "copy_data",
        mode="plain",
    )
    @classmethod
    def cast_to_copy_data(cls, value: bool | None) -> bool | None:
        if value in (None, "", "None"):
            return None
        else:
            return convert_str_to_bool(value)

    @field_validator(
        "url",
        mode="plain",
    )
    @classmethod
    def cast_to_url(cls, value: str | DataUrl | None) -> DataUrl:
        if value is None:
            return value
        elif isinstance(value, DataUrl):
            return value
        else:
            if "path=" not in value:
                # by default we expect the user to only give an entry on the file. So only the dataset path
                return DataUrl(data_path=value)
            else:
                return DataUrl(path=value)

    @field_serializer(
        "url",
        when_used="always",
    )
    @classmethod
    def serialize_data_url(cls, value: DataUrl | None) -> str:
        if value is None:
            return ""
        elif value.file_path() is None:
            return value.data_path()
        else:
            return value.path()

    @staticmethod
    def frm_str(input_str: str) -> FrameGroup:
        """
        Create an instance of FrameGroup from it string representation.
        """
        if not isinstance(input_str, str):
            raise TypeError(f"{input_str} should be a string")

        input_str = remove_parenthesis_or_brackets(input_str)
        elmts = input_str.split(",")
        elmts = filter(None, [elmt.lstrip(" ").rstrip(" ") for elmt in elmts])

        cst_inputs = {}

        for elmt in elmts:
            try:
                info_type, value = FrameGroup._treat_elmt(elmt)
            except ValueError:
                url_example = DataUrl(
                    file_path="/path/to/my/file/file.h5",
                    data_path="/data/path",
                    scheme="h5py",
                )
                _example_frame = FrameGroup(
                    url=url_example, copy_data=True, frame_type="projection"
                )
                err_msg = (
                    f"Unable to interpret string ('{elmt}'). Please insure this is a "
                    "either a frame type, a boolean for copy or an entry "
                    "(DataUrl).\n "
                    "Please prefix the value by the information type like: "
                    f"{_example_frame}. Invalid element is {input_str}"
                )
                raise ValueError(err_msg)
            else:
                cst_inputs[info_type] = value

        inputs = {}
        if FrameGroup.Info.FRAME_TYPE not in cst_inputs:
            raise ValueError(f"Unable to find frame type from {input_str}")
        else:
            inputs["frame_type"] = cst_inputs[FrameGroup.Info.FRAME_TYPE]
        if FrameGroup.Info.URL_ENTRY not in cst_inputs:
            raise ValueError(f"Unable to find entry from {input_str}")
        else:
            inputs["url"] = cst_inputs[FrameGroup.Info.URL_ENTRY]
        if FrameGroup.Info.COPY in cst_inputs:
            inputs["copy_data"] = cst_inputs[FrameGroup.Info.COPY]

        if "copy_data" in inputs:
            inputs["copy_data"] = inputs["copy_data"]

        return FrameGroup(**inputs)

    @staticmethod
    def _treat_elmt(elmt: str) -> tuple[str, str]:
        assert isinstance(elmt, str)
        # try to interpret it as a Frame group info
        for info in FrameGroup.Info:
            key = f"{info.value}="
            if elmt.startswith(key):
                return info, filter_str_def(elmt.replace(key, "", 1))

        # try to interpret it as an acquisition step
        elmt = filter_str_def(elmt)
        try:
            acquisition_step = AcquisitionStep.from_value(elmt)
        except ValueError:
            pass
        else:
            return FrameGroup.Info.FRAME_TYPE, acquisition_step

        # try to interpret it as a boolean
        # is this a copy element
        if elmt.startswith(("copy=", "copy_data=")):
            return FrameGroup.Info.COPY, elmt.split("=")[1]

        if elmt in ("True", "true"):
            return FrameGroup.Info.COPY, True
        if elmt in ("False", "false"):
            return FrameGroup.Info.COPY, False

        try:
            elmt = filter_str_def(elmt)
            DataUrl(path=elmt)
        except ValueError:
            pass
        else:
            return FrameGroup.Info.URL_ENTRY, elmt

        raise ValueError

    def __str__(self) -> str:
        return self.str_representation(
            only_data_path=False, with_copy=True, with_prefix_key=True
        )

    def str_representation(
        self, only_data_path: bool, with_copy: bool, with_prefix_key: bool
    ) -> str:
        """
        Util function to print the possible input string for this FrameGroup.

        :param only_data_path: if True consider the input file frame group is
                               contained in the input file and the string
                               representing the url can be only the data path
        :param with_copy: if true display the copy information
        :param with_prefix_key: if true provide the string with the keys as
                                prefix (frame_type=XXX, copy=...)
        """
        if self.url is None:
            url_str = ""
        elif only_data_path or self.url.file_path() is None:
            url_str = self.url.data_path()
        else:
            url_str = self.url.path()
        if with_prefix_key:
            if with_copy:
                return "({ft_key}={frame_type}, {url_key}={url}, {copy_key}={copy})".format(
                    ft_key=self.Info.FRAME_TYPE.value,
                    frame_type=self.frame_type.value,
                    url_key=self.Info.URL_ENTRY.value,
                    url=url_str,
                    copy_key=self.Info.COPY.value,
                    copy=self.copy_data,
                )
            else:
                return "({ft_key}={frame_type}, {url_key}={url})".format(
                    ft_key=self.Info.FRAME_TYPE.value,
                    frame_type=self.frame_type.value,
                    url_key=self.Info.URL_ENTRY.value,
                    url=url_str,
                )
        else:
            if with_copy:
                return "({frame_type}, {url}, {copy})".format(
                    frame_type=self.frame_type.value,
                    url=url_str,
                    copy=self.copy_data,
                )
            else:
                return "({frame_type}, {url})".format(
                    frame_type=self.frame_type.value,
                    url=url_str,
                )


def filter_acqui_frame_type(
    init: FrameGroup, sequences: tuple, frame_type: AcquisitionStep
) -> tuple:
    """compute the list of urls representing projections
    from init until the next Initialization step

    :param init: frame group creating the beginning of the
                            acquisition sequence
    :param sequences: list of FrameGroup representing the sequence
    :param frame_type: type of frame to filer (cannot be Initialization step)
    """
    frame_type = AcquisitionStep.from_value(frame_type)
    if frame_type is AcquisitionStep.INITIALIZATION:
        raise ValueError(f"{AcquisitionStep.INITIALIZATION.value} is not handled")
    if init not in sequences:
        raise ValueError(f"{init} cannot be find in the provided sequence")

    frame_types = [frm_grp.frame_type for frm_grp in sequences]
    current_acqui_idx = sequences.index(init)
    if len(sequences) == current_acqui_idx - 1:
        # in case the initialization sequence is the last element of the
        # sequence (if people make strange stuff...)
        return ()
    sequence_target = sequences[current_acqui_idx + 1 :]
    frame_types = frame_types[current_acqui_idx + 1 :]
    try:
        next_acqui = frame_types.index(AcquisitionStep.INITIALIZATION) - 1
    except ValueError:
        next_acqui = -1
    sequence_target = sequence_target[:next_acqui]
    filter_fct = lambda a: a.frame_type is frame_type
    return tuple(filter(filter_fct, sequence_target))



# coding: utf-8

"""
module to convert fluo-tomo files (after PyMCA fit, tif files) to (nexus tomo compliant) .nx
"""

from __future__ import annotations

import pint
import logging
import os

from tqdm import tqdm

from .fluoscan import FluoTomoScan2D, FluoTomoScan3D
from .blissfluoscan import BlissFluoTomoScan3D

from nxtomomill.models.fluo2nx import Fluo2nxModel

from nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill import utils

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)

__all__ = ["from_fluo_to_nx", "from_blissfluo_to_nx"]


def from_fluo_to_nx(configuration: Fluo2nxModel, progress: tqdm | None = None) -> tuple:
    """
    Converts an fluo-tomo tiff files to a nexus file.
    For now duplicates data.

    :param configuration: configuration to use to process the data
    :param progress: if provided then will be updated with conversion progress
    :return: (nexus_file, entry)
    """
    if configuration.input_folder is None:
        raise ValueError("input_folder should be provided")
    if not os.path.isdir(configuration.input_folder):
        raise OSError(f"{configuration.input_folder} is not a valid folder path")

    if configuration.output_file is None:
        raise ValueError("output_file should be provided")

    if configuration.detector_names is None:
        raise ValueError("Detector names should be provided.")

    fileout_h5 = utils.get_file_name(
        file_name=configuration.output_file,
        extension=configuration.file_extension,
        check=True,
    )

    if configuration.dimension == 2:
        scan = FluoTomoScan2D(
            scan=configuration.input_folder,
            dataset_basename=configuration.dataset_basename,
            detectors=configuration.detector_names,
        )
    elif configuration.dimension == 3:
        scan = FluoTomoScan3D(
            scan=configuration.input_folder,
            dataset_basename=configuration.dataset_basename,
            detectors=configuration.detector_names,
        )
    else:
        raise ValueError(f"Dimension should be 2 or 3 not {configuration.dimension}.")

    if progress is not None:
        progress.set_description("fluo2nx")
        progress.total = len(scan.el_lines)
        progress.n = 0

    _logger.info(f"Fluo lines preset in dataset are {scan.el_lines}")

    entry_list = []
    for element, lines in scan.el_lines.items():
        if progress is not None:
            progress.set_postfix_str(f"elmt - {element}")
            line_progress = tqdm(
                desc=f"elmt: {element} - line: ", position=1, leave=False
            )
            line_progress.total = len(lines)
        else:
            line_progress = None
        for i_line, line in enumerate(lines):
            if line_progress is not None:
                line_progress.set_postfix_str(f"elmt: {element} - line: {line}")

            for det in scan.detectors:
                elmt_line_data = scan.load_data(det, element=element, line_ind=i_line)
                if configuration.dimension == 2:
                    elmt_line_data = elmt_line_data.swapaxes(0, 1).copy()
                # Otherwise, it's 3D case, and the structure of elmt_line_data is OK.
                #
                my_nxtomo = NXtomo()
                my_nxtomo.instrument.detector.data = elmt_line_data
                my_nxtomo.instrument.detector.image_key_control = [
                    ImageKey.PROJECTION
                ] * elmt_line_data.shape[0]
                my_nxtomo.sample.rotation_angle = scan.rot_angles_deg * _ureg.degree
                my_nxtomo.instrument.detector.x_pixel_size = scan.pixel_size * _ureg(
                    "m"
                )
                my_nxtomo.instrument.detector.y_pixel_size = scan.pixel_size * _ureg(
                    "m"
                )

                # define a value to sample-detector and source-sample distance. To be set to the real value in the future
                my_nxtomo.instrument.detector.distance = 1.0 * _ureg.meter
                my_nxtomo.instrument.source.distance = 1.0 * _ureg.meter
                my_nxtomo.energy = scan.energy * _ureg.keV

                data_path = f"{det}_{element}_{line}"
                my_nxtomo.save(
                    file_path=fileout_h5,
                    data_path=data_path,
                    overwrite=configuration.overwrite,
                )
                entry_list.append((fileout_h5, data_path))
            if line_progress is not None:
                line_progress.update()
        if progress is not None:
            progress.update()
    return tuple(entry_list)


def from_blissfluo_to_nx(
    configuration: Fluo2nxModel, progress: tqdm | None = None
) -> tuple:
    """
    Converts an Bliss fluo-tomo h5 file to a nexus file.
    For now duplicates data.

    :param configuration: configuration to use to process the data
    :param progress: if provided then will be updated with conversion progress
    :return: (nexus_file, entry)
    """
    if configuration.general_section.ewoksfluo_filename is None:
        raise ValueError("ewoksfluo_filename should be provided")
    if not os.path.isfile(configuration.general_section.ewoksfluo_filename):
        raise FileNotFoundError(
            f"{configuration.general_section.ewoksfluo_filename} is not a valid filename path"
        )

    if configuration.general_section.output_file is None:
        raise ValueError("output_file should be provided")

    if configuration.general_section.detector_names is None:
        raise ValueError("Detector names should be provided.")

    fileout_h5 = utils.get_file_name(
        file_name=configuration.general_section.output_file,
        extension=configuration.general_section.file_extension,
        check=True,
    )

    if configuration.general_section.dimension == 2:
        raise NotImplementedError("Only available for 3D at the time")
    elif configuration.general_section.dimension == 3:
        scan = BlissFluoTomoScan3D(
            ewoksfluo_filename=configuration.general_section.ewoksfluo_filename,
            detectors=configuration.general_section.detector_names,
        )
    else:
        raise ValueError(
            f"Dimension should be 2 or 3 not {configuration.general_section.dimension}."
        )

    if progress is not None:
        progress.set_description("blissfluo2nx")
        progress.total = len(scan.lines) * len(scan.detectors)
        progress.n = 0

    _logger.info(f"Fluo lines preset in dataset are {scan.lines}")

    entry_list = []
    for line in scan.lines:
        for det in scan.detectors:
            line_data = scan.load_data(det, line=line)
            #
            my_nxtomo = NXtomo()
            my_nxtomo.instrument.detector.data = line_data
            my_nxtomo.instrument.detector.image_key_control = [
                ImageKey.PROJECTION
            ] * line_data.shape[0]
            my_nxtomo.sample.rotation_angle = scan.rot_angles_deg
            # TODO: should the pixel_size be in a more readable unit (convert to micrometer ?)
            my_nxtomo.instrument.detector.x_pixel_size = scan.pixel_size.to(_ureg.meter)
            my_nxtomo.instrument.detector.y_pixel_size = scan.pixel_size.to(_ureg.meter)

            # define a value to sample-detector and source-sample distance. To be set to the real value in the future
            my_nxtomo.instrument.detector.distance = 1.0 * _ureg.meter
            my_nxtomo.instrument.source.distance = 1.0 * _ureg.meter
            my_nxtomo.instrument.source.name = configuration.source_section.source_name
            my_nxtomo.instrument.source.type = configuration.source_section.source_type
            my_nxtomo.instrument.source.probe = (
                configuration.source_section.source_probe
            )

            my_nxtomo.instrument.name = configuration.instrument_section.instrument_name
            my_nxtomo.energy = scan.energy

            data_path = f"{det}_{line}"
            my_nxtomo.save(
                file_path=fileout_h5,
                data_path=data_path,
                overwrite=configuration.general_section.overwrite,
            )
            entry_list.append((fileout_h5, data_path))
        if progress is not None:
            progress.update()
    return tuple(entry_list)



# coding: utf-8

import os
import tempfile

import pytest
from silx.io.url import DataUrl

from tomoscan.io import HDF5File
from nxtomomill.converter.hdf5.acquisition.baseacquisition import (
    BaseAcquisition,
    get_dataset_name_from_motor,
)
from nxtomomill.io.config import TomoHDF5Config


def test_BaseAquisition():
    """simple test of the BaseAcquisition class"""

    with tempfile.TemporaryDirectory() as folder:
        file_path = os.path.join(folder, "test.h5")
        with HDF5File(file_path, mode="w") as h5f:
            h5f["/data/toto/dataset"] = 12

        url = DataUrl(file_path=file_path, data_path="/data/toto", scheme="silx")
        std_acq = BaseAcquisition(
            root_url=url,
            configuration=TomoHDF5Config(),
            detector_sel_callback=None,
            start_index=0,
        )
        with std_acq.read_entry() as entry:
            assert "dataset" in entry


def test_get_dataset_name_from_motor():
    """test get_dataset_name_from_motor function"""
    set_1 = ["rotation", "test1", "alias"]
    assert get_dataset_name_from_motor(set_1, "rotation") == "test1"
    assert get_dataset_name_from_motor(set_1, "my motor") is None

    set_2 = ["rotation", "test1", "alias", "x translation", "m2", "test1"]
    assert get_dataset_name_from_motor(set_2, "rotation") == "test1"
    assert get_dataset_name_from_motor(set_2, "x translation") == "m2"

    set_3 = ["rotation"]
    with pytest.raises(ValueError):
        get_dataset_name_from_motor(set_3, "rotation")



# coding: utf-8

import unittest

from nxtomomill.models import utils


class TestConvertStrToTuple(unittest.TestCase):
    """
    test convert_str_to_tuple function
    """

    def testStr1(self):
        self.assertEqual(utils.convert_str_to_tuple("toto, tata"), ("toto", "tata"))

    def testStr2(self):
        self.assertEqual(
            utils.convert_str_to_tuple("'toto', \"tata\""), ("toto", "tata")
        )

    def testStr3(self):
        self.assertEqual(utils.convert_str_to_tuple("test"), ("test",))

    def testStr4(self):
        self.assertEqual(
            utils.convert_str_to_tuple("(this is a test)"), ("this is a test",)
        )

    def testStr5(self):
        self.assertEqual(
            utils.convert_str_to_tuple("(this is a test, 'and another one')"),
            ("this is a test", "and another one"),
        )


class TestIsUrlPath(unittest.TestCase):
    """test the is_url function"""

    def test_invalid_url_1(self):
        self.assertFalse(utils.is_url_path("/toto/tata"))

    def test_invalid_url_2(self):
        self.assertFalse(utils.is_url_path("tata"))

    def test_valid_url_1(self):
        self.assertTrue(utils.is_url_path("silx:///data/image.h5?path=/scan_0/data"))

    def test_valid_url_2(self):
        self.assertTrue(utils.is_url_path("silx:///data/image.edf"))

    def test_valid_url_3(self):
        self.assertTrue(utils.is_url_path("silx://image.h5"))

    def test_valid_url_4(self):
        self.assertTrue(
            utils.is_url_path("silx:///data/image.h5?path=/scan_0/data&slice=1,5")
        )



import os
import numpy
import pytest
import pint
from tomoscan.io import HDF5File, get_swmr_mode
from nxtomo.application.nxtomo import NXtomo
from nxtomomill.app.split_nxfile import split

_ureg = pint.get_application_registry()


@pytest.mark.parametrize("duplicate_data", (True, False))
def test_split_nxfile(tmp_path, duplicate_data):
    """
    test execution of split_nxfile
    One file contains
    """
    nx_tomo = NXtomo()
    n_frames = 20
    nx_tomo.instrument.detector.data = numpy.random.random(
        100 * 100 * n_frames
    ).reshape([n_frames, 100, 100])
    nx_tomo.sample.rotation_angle = [0, 12] * _ureg.degree

    output_dir = tmp_path / "test_split_nx_file"
    output_dir.mkdir()
    output_dir = str(output_dir)
    output_file = os.path.join(output_dir, "input_file.nx")
    nx_tomo.save(output_file, "entry0000")
    nx_tomo.save(output_file, "entry0001")
    nx_tomo.save(output_file, "entry0002")

    # test split
    with pytest.raises(ValueError):
        # make sure if no '{index}' or '{entry_name}' are provided this raises an error
        split(
            output_file,
            output_file_pattern=os.sep.join([output_dir, "output_dir_1", "part_.nx"]),
        )

    split(
        output_file,
        output_file_pattern=os.sep.join(
            [output_dir, "output_dir_1", "part_{index}.nx"]
        ),
        duplicate_data=duplicate_data,
    )
    for i_file in range(3):
        assert os.path.exists(
            tmp_path / f"test_split_nx_file/output_dir_1/part_{i_file}.nx"
        )

    split(
        output_file,
        output_file_pattern=os.sep.join(
            [output_dir, "output_dir_2", "{entry_name}.nx"]
        ),
        duplicate_data=duplicate_data,
    )
    for output in ("entry0000", "entry0001", "entry0002"):
        assert os.path.exists(tmp_path / f"test_split_nx_file/output_dir_2/{output}.nx")

    # test overwrite
    with pytest.raises(KeyError):
        split(
            output_file,
            output_file_pattern=os.sep.join(
                [output_dir, "output_dir_2", "{entry_name}.nx"]
            ),
            duplicate_data=duplicate_data,
        )

    with HDF5File(
        tmp_path / "test_split_nx_file/output_dir_2/entry0002.nx",
        mode="r",
        swmr=get_swmr_mode(),
    ) as h5f:
        assert "entry0002" in h5f.keys()
        assert "entry0000" not in h5f.keys()



from __future__ import annotations

from nxtomomill.settings import Tomo
from nxtomomill.models.utils import convert_str_to_tuple
from nxtomomill.utils.io import deprecated
from nxtomomill.models.utils import filter_str_def

from pydantic import BaseModel, Field, field_validator, ConfigDict


class DarkAndFlatSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    dark_names_prefix: tuple[str, ...] = Field(
        default=Tomo.EDF.DARK_NAMES, description="Prefixes of dark field file(s)"
    )
    flat_names_prefix: tuple[str, ...] = Field(
        default=Tomo.EDF.REFS_NAMES, description="Prefixes of flat field file(s)"
    )

    @field_validator(
        "dark_names_prefix",
        "flat_names_prefix",
        mode="plain",
    )
    @classmethod
    def cast_to_tuple(cls, value: str | tuple[str, ...]) -> tuple[str,]:
        if isinstance(value, str):
            value = filter_str_def(value)
        return convert_str_to_tuple(value) or ()

    @property
    @deprecated(since_version="2.0", reason="renamed", replacement="dark_names_prefix")
    def dark_names(self):
        return self.dark_names_prefix

    @dark_names.setter
    @deprecated(since_version="2.0", reason="renamed", replacement="dark_names_prefix")
    def dark_names(self, value):
        self.dark_names_prefix = value

    @property
    @deprecated(since_version="2.0", reason="renamed", replacement="flat_names_prefix")
    def flat_names(self):
        return self.flat_names_prefix

    @flat_names.setter
    @deprecated(since_version="2.0", reason="renamed", replacement="flat_names_prefix")
    def flat_names(self, value):
        self.flat_names_prefix = value



from __future__ import annotations

import os
import pytest
import logging
import importlib.resources as importlib_resources

from silx.io.url import DataUrl

from nxtomo.nxobject.nxdetector import FieldOfView

from nxtomomill import settings
from nxtomomill.utils import FileExtension
from nxtomomill.models.h52nx import H52nxModel, generate_default_h5_config
from nxtomomill.models.h52nx.general_section import GeneralSection
from nxtomomill.models.h52nx.keys_section import KeysSection
from nxtomomill.models.h52nx.entries_and_title_section import (
    EntriesAndTitlesSection,
)
from nxtomomill.models.h52nx.frame_type_section import (
    FrameTypeSection,
    FrameGroup,
)
from nxtomomill.models.h52nx.multitomo_section import MultiTomoSection
from nxtomomill.models.h52nx.extra_params_section import (
    ExtraParamsSection,
)

from nxtomomill.tests.resources import h52nx_config_files as _h52nx_config_files
from configparser import ConfigParser


def test_moving_to_pydantic_default_config_file():
    """
    test that the default nxtomomill config generated before moving to pydantic is properly interpreted.
    This ignores comments.
    """
    resources = importlib_resources.files(_h52nx_config_files.__name__)
    ref_file = os.path.join(resources, "default_file_before_moving_to_pydantic.cfg")
    assert os.path.exists(ref_file)
    txt_parser = ConfigParser(allow_no_value=True)
    txt_parser.read(ref_file)

    old_config = H52nxModel(
        general_section=txt_parser["GENERAL_SECTION"],
        keys_section=txt_parser["KEYS_SECTION"],
        entries_and_titles_section=txt_parser["ENTRIES_AND_TITLES_SECTION"],
        frame_type_section=txt_parser["FRAME_TYPE_SECTION"],
        multitomo_section=txt_parser["MULTITOMO_SECTION"],
        extra_params_section=txt_parser["EXTRA_PARAMS_SECTION"],
    ).to_nested_model()
    old_dict = {key.upper(): value for key, value in old_config.model_dump().items()}

    new_dict = generate_default_h5_config()
    assert new_dict == old_dict


def test_moving_to_pydantic_modified_config_file():
    """
    test that an old nxtomomill h5-config file generated before moving to pydantic is properly interpreted.
    This ignores any comments.
    """
    resources = importlib_resources.files(_h52nx_config_files.__name__)
    ref_file = os.path.join(resources, "fully_modified_file_before_moving.cfg")
    assert os.path.exists(ref_file)

    config = H52nxModel.from_cfg_file(ref_file)
    # general section
    assert config.input_file == "input_file.h5"
    assert config.output_file == "output_file.nx"
    assert config.overwrite is True
    assert config.file_extension is FileExtension.H5
    assert config.log_level is logging.DEBUG
    assert config.raises_error is True
    assert config.no_input is True
    assert config.single_file is True
    assert config.no_master_file is False
    assert config.ignore_bliss_tomo_config is True
    assert config.field_of_view is FieldOfView.FULL
    assert config.create_control_data is False
    # keys section
    assert config.valid_camera_names == ("my_camera",)
    assert config.rotation_angle_keys == ("rotm",)
    assert config.sample_x_keys == ("samx",)
    assert config.sample_y_keys == ("samy",)
    assert config.translation_y_keys == ("yrot",)
    assert config.translation_z_keys == ("zrot",)
    assert config.diode_keys == ("fpico",)
    assert config.exposure_time_keys == ("exposure_time",)
    assert config.sample_x_pixel_size_keys == ("technique/optic/my_sample_pixel_size",)
    assert config.sample_y_pixel_size_keys == ("technique/optic/my_sample_pixel_size2",)
    assert config.detector_x_pixel_size_keys == ("technique/scan/detector_pixel_size",)
    assert config.detector_y_pixel_size_keys == ("technique/scan/detector_pixel_size2",)
    assert config.sample_detector_distance_keys == (
        "technique/scan/my_sample_detector_distance",
    )
    assert config.source_sample_distance_keys == (
        "technique/scan/source_my_sample_distance",
    )
    # entries and titles section
    assert {url.path() for url in config.entries} == {
        DataUrl(data_path="/1.1", scheme="silx").path(),
        DataUrl(data_path="/3.1", scheme="silx").path(),
    }
    assert {url.path() for url in config.sub_entries_to_ignore} == {
        DataUrl(data_path="/2.1", scheme="silx").path()
    }
    assert config.init_titles == ("my_tomo:basic",)
    assert config.zseries_init_titles == ("my_tomo:zseries",)
    assert config.multitomo_init_titles == ("my_tomo:pcotomo",)
    assert config.dark_titles == ("my dark images",)
    assert config.flat_titles == ("my flat",)
    assert config.projection_titles == ("my projections",)
    assert config.alignment_titles == ("my static images",)
    # frame type section
    assert config.data_scans == ()
    assert config.default_data_copy is True
    # multitomo section
    assert config.start_angle_offset_in_degree == 24.2
    assert config.n_nxtomo == 100
    assert config.angle_interval_in_degree == 180
    assert config.shift_angles is True
    # extra params
    assert config.energy_kev == 12.5
    assert config.y_detector_pixel_size_m == 2.3e-6


def test_moving_to_pydantic_file_with_data_scans():
    resources = importlib_resources.files(_h52nx_config_files.__name__)
    ref_file = os.path.join(resources, "config_file_with_data_scans.cfg")
    assert os.path.exists(ref_file)
    txt_parser = ConfigParser(allow_no_value=True)
    txt_parser.read(ref_file)

    old_config = H52nxModel(
        general_section=txt_parser["GENERAL_SECTION"],
        keys_section=txt_parser["KEYS_SECTION"],
        entries_and_titles_section=txt_parser["ENTRIES_AND_TITLES_SECTION"],
        frame_type_section=txt_parser["FRAME_TYPE_SECTION"],
        multitomo_section=txt_parser["MULTITOMO_SECTION"],
        extra_params_section=txt_parser["EXTRA_PARAMS_SECTION"],
    )
    old_dict = {key.upper(): value for key, value in old_config.model_dump().items()}

    config = H52nxModel(
        general_section=GeneralSection(),
        keys_section=KeysSection(),
        entries_and_titles_section=EntriesAndTitlesSection(),
        frame_type_section=FrameTypeSection(
            data_scans=(
                FrameGroup(
                    url="silx:///path/to/file?/path/to/scan/node",
                    frame_type="projections",
                ),
                FrameGroup(
                    url="/path_relative_to_file",
                    frame_type="darks",
                    copy_data=True,
                ),
            )
        ),
        multitomo_section=MultiTomoSection(),
        extra_params_section=ExtraParamsSection(),
    )
    new_dict = {key.upper(): value for key, value in config.model_dump().items()}
    assert new_dict == old_dict


def test_generate_default_hdf5_config():
    """
    Insure we can generate a default configuration
    """
    config = H52nxModel()
    config.input_file = "toto.h5"
    config.output_file = "toto.nx"
    output = config.to_dict()

    assert isinstance(output, dict)
    # check titles values
    titles_dict = output["ENTRIES_AND_TITLES_SECTION"]
    assert titles_dict["init_titles"] == settings.Tomo.H5.INIT_TITLES
    assert titles_dict["zseries_init_titles"] == settings.Tomo.H5.ZSERIE_INIT_TITLES
    assert titles_dict["projection_titles"] == settings.Tomo.H5.PROJ_TITLES
    assert titles_dict["flat_titles"] == settings.Tomo.H5.FLAT_TITLES
    assert titles_dict["dark_titles"] == settings.Tomo.H5.DARK_TITLES
    assert titles_dict["alignment_titles"] == settings.Tomo.H5.ALIGNMENT_TITLES
    # check sample pixel size
    keys_dict = output["KEYS_SECTION"]
    assert (
        keys_dict["sample_x_pixel_size_keys"]
        == settings.Tomo.H5.SAMPLE_X_PIXEL_SIZE_KEYS
    )
    assert (
        keys_dict["sample_y_pixel_size_keys"]
        == settings.Tomo.H5.SAMPLE_Y_PIXEL_SIZE_KEYS
    )
    assert (
        keys_dict["detector_x_pixel_size_keys"]
        == settings.Tomo.H5.DETECTOR_X_PIXEL_SIZE_KEYS
    )
    assert (
        keys_dict["detector_y_pixel_size_keys"]
        == settings.Tomo.H5.DETECTOR_Y_PIXEL_SIZE_KEYS
    )
    # check translation
    assert keys_dict["sample_x_keys"] == settings.Tomo.H5.SAMPLE_X_KEYS
    assert keys_dict["sample_y_keys"] == settings.Tomo.H5.SAMPLE_Y_KEYS
    assert keys_dict["translation_z_keys"] == settings.Tomo.H5.TRANSLATION_Z_KEYS
    # others
    if len(settings.Tomo.H5.VALID_CAMERA_NAMES) == 0:
        assert keys_dict["valid_camera_names"] == ()
    else:
        assert keys_dict["valid_camera_names"] == settings.Tomo.H5.VALID_CAMERA_NAMES
    assert keys_dict["rotation_angle_keys"] == settings.Tomo.H5.ROT_ANGLE_KEYS
    assert keys_dict["diode_keys"] == settings.Tomo.H5.DIODE_KEYS
    assert keys_dict["translation_y_keys"] == settings.Tomo.H5.TRANSLATION_Y_KEYS
    assert keys_dict["exposure_time_keys"] == settings.Tomo.H5.ACQ_EXPO_TIME_KEYS

    # check input and output file
    general_information = output["GENERAL_SECTION"]
    assert general_information["input_file"] == "toto.h5"
    assert general_information["output_file"] == "toto.nx"


def test_hdf5config_to_dict():
    """test the `to_dict` function"""
    output_dict = H52nxModel().to_nested_model().model_dump()
    # check sections
    for section in (
        "general_section",
        "keys_section",
        "extra_params_section",
        "frame_type_section",
        "entries_and_titles_section",
    ):
        assert section in output_dict
    # check titles keys
    for key in (
        "alignment_titles",
        "projection_titles",
        "zseries_init_titles",
        "init_titles",
        "flat_titles",
        "dark_titles",
    ):
        assert key in output_dict["entries_and_titles_section"]
    # check sample pixel size
    for key in (
        "sample_x_pixel_size_keys",
        "sample_y_pixel_size_keys",
        "detector_x_pixel_size_keys",
        "detector_y_pixel_size_keys",
    ):
        assert key in output_dict["keys_section"]
    # translation keys
    for key in (
        "sample_x_keys",
        "sample_y_keys",
        "translation_z_keys",
    ):
        assert key in output_dict["keys_section"]
    # others
    for key in (
        "valid_camera_names",
        "rotation_angle_keys",
        "translation_y_keys",
        "diode_keys",
        "exposure_time_keys",
    ):
        assert key in output_dict["keys_section"]


def test_hdf5config_from_dict():
    """test the `from_dict` function"""
    valid_camera_names = ("frelon", "totocam")
    alignment_titles = ("this is an alignment",)
    sample_x_keys = ("tx", "x")
    config = H52nxModel.from_dict(
        {
            "KEYS_SECTION": {
                "valid_camera_names": valid_camera_names,
                "sample_x_keys": sample_x_keys,
            },
            "ENTRIES_AND_TITLES_SECTION": {"alignment_titles": alignment_titles},
        }
    )
    assert config.valid_camera_names == valid_camera_names
    assert config.alignment_titles == alignment_titles
    assert config.sample_x_keys == sample_x_keys


def test_hdf5config_from_dict_lowercase():
    """Test `from_dict` method with lower case sections"""
    ref_dict = H52nxModel().to_dict()
    lower_dict = {k.lower(): v for k, v in ref_dict.items()}

    config = H52nxModel.from_dict(lower_dict)
    assert config.to_dict() == ref_dict


def test_hdf5config_raises_errors():
    """
    Insure a type error is raised if an invalid type is passed to the
    HDF5Config class
    :return:
    """
    with pytest.raises(TypeError):
        H52nxModel.from_dict({"ENTRIES_AND_TITLES_SECTION": {"dark_titles": 1213}})


def test_hdf5config_to_and_from_cfg_file(tmp_path):
    """
    Insure we can dump the configuration to a .cfg file and that we
    can read it back
    """
    file_path = os.path.join(tmp_path, "output_file.cfg")
    input_config = H52nxModel()
    input_config.to_cfg_file(file_path)
    assert os.path.exists(file_path)
    loaded_config = H52nxModel.from_cfg_file(file_path=file_path)
    assert isinstance(loaded_config, H52nxModel)



# coding: utf-8

from __future__ import annotations

from pydantic import BaseModel, ConfigDict
from configparser import ConfigParser

from ..base.NestedModelBase import NestedModelBase
from ..base.FrmFlatToNestedBase import FrmFlatToNestedBase
from ..base.instrument_section import InstrumentSection
from .general_section import GeneralSection
from .keys_section import KeysSection
from .entries_and_title_section import EntriesAndTitlesSection
from .frame_type_section import FrameTypeSection
from .multitomo_section import MultiTomoSection
from .extra_params_section import ExtraParamsSection
from nxtomomill.utils.io import deprecated_warning, deprecated

__all__ = [
    "H52nxModel",
    "generate_default_h5_config",
]


class H52nxModel(
    FrmFlatToNestedBase,
    GeneralSection,
    KeysSection,
    EntriesAndTitlesSection,
    FrameTypeSection,
    MultiTomoSection,
    ExtraParamsSection,
    InstrumentSection,
):
    """Configuration file to run a conversion from bliss-hdf5 file to NXtomo format"""

    model_config: ConfigDict = ConfigDict(validate_assignment=True)

    def to_nested_model(self) -> _NestedTomoHDF5Config:
        return _NestedTomoHDF5Config(
            general_section=GeneralSection(
                output_file=self.output_file,
                overwrite=self.overwrite,
                log_level=self.log_level,
                input_file=self.input_file,
                file_extension=self.file_extension,
                raises_error=self.raises_error,
                no_input=self.no_input,
                single_file=self.single_file,
                no_master_file=self.no_master_file,
                ignore_bliss_tomo_config=self.ignore_bliss_tomo_config,
                field_of_view=self.field_of_view,
                rotation_is_clockwise=self.rotation_is_clockwise,
                create_control_data=self.create_control_data,
                check_tomo_n=self.check_tomo_n,
                mechanical_lr_flip=self.mechanical_lr_flip,
                mechanical_ud_flip=self.mechanical_ud_flip,
            ),
            keys_section=KeysSection(
                valid_camera_names=self.valid_camera_names,
                rotation_angle_keys=self.rotation_angle_keys,
                sample_x_keys=self.sample_x_keys,
                sample_y_keys=self.sample_y_keys,
                translation_y_keys=self.translation_y_keys,
                translation_z_keys=self.translation_z_keys,
                diode_keys=self.diode_keys,
                exposure_time_keys=self.exposure_time_keys,
                sample_x_pixel_size_keys=self.sample_x_pixel_size_keys,
                sample_y_pixel_size_keys=self.sample_y_pixel_size_keys,
                detector_x_pixel_size_keys=self.detector_x_pixel_size_keys,
                detector_y_pixel_size_keys=self.detector_y_pixel_size_keys,
                sample_detector_distance_keys=self.sample_detector_distance_keys,
                source_sample_distance_keys=self.source_sample_distance_keys,
                machine_current_keys=self.machine_current_keys,
            ),
            entries_and_titles_section=EntriesAndTitlesSection(
                entries=self.entries,
                sub_entries_to_ignore=self.sub_entries_to_ignore,
                init_titles=self.init_titles,
                zseries_init_titles=self.zseries_init_titles,
                multitomo_init_titles=self.multitomo_init_titles,
                dark_titles=self.dark_titles,
                flat_titles=self.flat_titles,
                projection_titles=self.projection_titles,
                alignment_titles=self.alignment_titles,
                back_and_forth_init_titles=self.back_and_forth_init_titles,
            ),
            frame_type_section=FrameTypeSection(
                data_scans=self.data_scans,
                default_data_copy=self.default_data_copy,
            ),
            multitomo_section=MultiTomoSection(
                start_angle_offset_in_degree=self.start_angle_offset_in_degree,
                n_nxtomo=self.n_nxtomo,
                angle_interval_in_degree=self.angle_interval_in_degree,
                shift_angles=self.shift_angles,
            ),
            extra_params_section=ExtraParamsSection(
                energy_kev=self.energy_kev,
                x_sample_pixel_size_m=self.x_sample_pixel_size_m,
                y_sample_pixel_size_m=self.y_sample_pixel_size_m,
                x_detector_pixel_size_m=self.x_detector_pixel_size_m,
                y_detector_pixel_size_m=self.y_detector_pixel_size_m,
                detector_sample_distance_m=self.detector_sample_distance_m,
                source_sample_distance_m=self.source_sample_distance_m,
            ),
        )

    @property
    def is_using_titles(self) -> bool:
        return len(self.data_scans) == 0

    @property
    def is_using_urls(self) -> bool:
        """
        Return true if we want to use urls for darks, flats, projections
        instead of titles
        """
        return len(self.data_scans) > 0

    def clear_entries_and_subentries(self):
        self.entries = ()
        self.sub_entries_to_ignore = ()

    @classmethod
    def from_cfg_file(cls, file_path: str) -> H52nxModel:
        txt_parser = ConfigParser(allow_no_value=True)
        txt_parser.read(file_path)

        def get_section(name, default={}):
            if txt_parser.has_section(name):
                return txt_parser[name]
            else:
                return default

        return _NestedTomoHDF5Config(
            general_section=GeneralSection(**get_section("GENERAL_SECTION")),
            entries_and_titles_section=EntriesAndTitlesSection(
                **get_section("ENTRIES_AND_TITLES_SECTION")
            ),
            extra_params_section=ExtraParamsSection(
                **get_section("EXTRA_PARAMS_SECTION")
            ),
            frame_type_section=FrameTypeSection(**get_section("FRAME_TYPE_SECTION")),
            multitomo_section=MultiTomoSection(**get_section("MULTITOMO_SECTION")),
            keys_section=KeysSection(**get_section("KEYS_SECTION")),
        ).to_flatten_config()

    # deprecated function / properties from version 1

    @staticmethod
    def from_dict(dict_: dict) -> None:
        deprecated_warning(
            type_="function",
            name="from_dict",
            reason="replaced by pydantic 'model_dump' function",
            replacement="model_dump",
            since_version="2.0",
        )

        dict_ = {key.lower(): value for key, value in dict_.items()}
        config = _NestedTomoHDF5Config(**dict_)
        return config.to_flatten_config()

    @property
    @deprecated(since_version="2.0", reason="removed", replacement="")
    def bam_single_file(self):
        """This option has been removed. Usage of 'single_file' should be enough"""
        pass

    @bam_single_file.setter
    @deprecated(since_version="2.0", reason="removed", replacement="")
    def bam_single_file(self, bam: bool):
        """This option has been removed. Usage of 'single_file' should be enough"""
        pass


class _NestedTomoHDF5Config(BaseModel, NestedModelBase):
    """
    Nested model to dump the model to .cfg.
    It has historically sections when the model (config class is flat). But the parameters are the same
    """

    model_config = ConfigDict(str_to_upper=True)

    general_section: GeneralSection = GeneralSection()
    keys_section: KeysSection = KeysSection()
    entries_and_titles_section: EntriesAndTitlesSection = EntriesAndTitlesSection()
    frame_type_section: FrameTypeSection = FrameTypeSection()
    multitomo_section: MultiTomoSection = MultiTomoSection()
    extra_params_section: ExtraParamsSection = ExtraParamsSection()

    def to_flatten_config(self) -> H52nxModel:
        return H52nxModel(
            **self.general_section.model_dump(),
            **self.keys_section.model_dump(),
            **self.entries_and_titles_section.model_dump(),
            **self.frame_type_section.model_dump(),
            **self.multitomo_section.model_dump(),
            **self.extra_params_section.model_dump(),
        )


def generate_default_h5_config() -> dict:
    """generate a default configuration for converting hdf5 to nx"""
    config = H52nxModel().to_nested_model()
    return {key.upper(): value for key, value in config.model_dump().items()}
